#!/usr/bin/env python3
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import lc3
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as signal

matplotlib.use('QtAgg')

parser = argparse.ArgumentParser(description='LC3 Encoder')

parser.add_argument(
    '--frame_duration', help='Frame duration (ms)', type=float, default=10)

parser.add_argument(
    '--sample_rate', help='Sampling frequency (Hz)', type=float, default=48000)

parser.add_argument(
    '--hrmode', help='Enable high-resolution mode', action='store_true')

parser.add_argument(
    '--bitrate', help='Bitrate (bps)', type=int, default=96000)

parser.add_argument(
    '--libpath', help='LC3 Library path')

args = parser.parse_args()

# --- Setup encoder + decoder ---

fs = args.sample_rate

enc = lc3.Encoder(
    args.frame_duration, fs, hrmode=args.hrmode, libpath=args.libpath)
dec = lc3.Decoder(
    args.frame_duration, fs, hrmode=args.hrmode, libpath=args.libpath)

frame_len = enc.get_frame_samples()
delay_len = enc.get_delay_samples()

# --- Generate 10 seconds chirp ---

n = 10 * int(fs)
t = np.arange(n - (n % frame_len)) / fs
s = signal.chirp(t, f0=10, f1=fs/2, t1=t[-1], phi=-90, method='linear')

# --- Encoding / decoding loop ---

frame_size = enc.get_frame_bytes(args.bitrate)
bitrate = enc.resolve_bitrate(frame_size)

y = np.empty(len(s) + frame_len)

for i in range(0, len(s), frame_len):
    y[i:i+frame_len] = dec.decode(enc.encode(s[i:i+frame_len], frame_size))

y[len(s):] = dec.decode(enc.encode(np.zeros(frame_len), frame_size))
y = y[delay_len:len(s)+delay_len]

# --- Plot spectrograms ---

fig, (ax1, ax2) = plt.subplots(nrows=2, sharex=True)

NFFT = 512
for (ax, s) in [(ax1, s), (ax2, y)]:
    ax.specgram(s, Fs=fs, NFFT=NFFT, pad_to=4*NFFT, noverlap=NFFT//2,
                vmin=-160, vmax=0)

ax1.set_title('Input signal')
ax1.set_ylabel('Frequency (Hz)')

ax2.set_title(('Processed signal (%.1f kbps)' % (bitrate/1000)))
ax2.set_ylabel('Frequency (Hz)')

plt.show()



#!/usr/bin/env python3
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import matplotlib.pyplot as plt


def fast_exp2(x, t, p):

    p = p.astype(np.float32)
    x = x.astype(np.float32)

    m = ((x + 0.5/8) % (1/8)) - (0.5/8)
    e = int((x - m) * 8)

    y = ((((p[0]*m) + p[1])*m + p[2])*m + p[3])*m + p[4]
    y = y * 2**(e // 8) * t[e % 8]

    return y.astype(np.float32)

def approx_exp2():

    x = np.arange(0, 1/8, step=1e-6)
    p = np.polyfit(x, 2 ** x, 4)
    t = [ 2**(i/8) for i in range(8) ]

    x =  np.arange(-10, 10, step=1e-3)
    y = [ fast_exp2(x[i], t, p) for i in range(len(x)) ]

    e = np.abs(y - 2**x) / (2 ** x)

    print('{{ {:14.8e}, {:14.8e}, {:14.8e}, {:14.8e}, \n'
           '  {:14.8e}, {:14.8e}, {:14.8e}, {:14.8e}, '.format(*t))
    print('{{ {:14.8e}, {:14.8e}, {:14.8e}, {:14.8e}, {:14.8e} }}'.format(*p))
    print('Max relative error: ', np.max(e))
    print('Max RMS error: ', np.sqrt(np.mean(e ** 2)))

    if False:
        fig, (ax1, ax2) = plt.subplots(2)

        ax1.plot(x, 2**x, label='Reference')
        ax1.plot(x, y, label='Approximation')
        ax1.legend()

        ax2.plot(x, e, label='Relative Error')
        ax2.legend()

        plt.show()


def fast_log2(x, p):

    p = p.astype(np.float32)
    x = x.astype(np.float32)

    (x, e) = np.frexp(x)

    y = ((((p[0]*x) + p[1])*x + p[2])*x + p[3])*x + p[4]

    return (e ) + y.astype(np.float32)

def approx_log2():

    x = np.logspace(-1, 0, base=2, num=100)
    p = np.polyfit(x, np.log2(x), 4)

    x = np.logspace(-2, 5, num=10000)
    y = [ fast_log2(x[i], p) for i in range(len(x)) ]
    e = np.abs(y - np.log2(x))

    print('{{ {:14.8e}, {:14.8e}, {:14.8e}, {:14.8e}, {:14.8e} }}'
                .format(p[0], p[1], p[2], p[3], p[4]))
    print('Max absolute error: ', np.max(e))
    print('Max RMS error: ', np.sqrt(np.mean(e ** 2)))

    if False:
        fig, (ax1, ax2) = plt.subplots(2)

        ax1.plot(x, np.log2(x),  label='Reference')
        ax1.plot(x, y, label='Approximation')
        ax1.legend()

        ax2.plot(x, e, label = 'Absolute error')
        ax2.legend()

        plt.show()


def table_db_q16():

    k = 10 * np.log10(2)

    for i in range(32):
        a = k * np.log2(np.ldexp(32 + i  , -5)) - (i // 16) * (k/2)
        b = k * np.log2(np.ldexp(32 + i+1, -5)) - (i // 16) * (k/2)

        an = np.ldexp(a, 15) + 0.5
        bn = np.ldexp(b - a, 15) + 0.5
        print('{{ {:5d}, {:4d} }},'
            .format(int(np.ldexp(a, 15) + 0.5),
                    int(np.ldexp(b - a, 15) + 0.5)),
            end = ' ' if i % 4 < 3 else '\n')


if __name__ == '__main__':

    print('\n--- Approximation of 2^n ---')
    approx_exp2()

    print('\n--- Approximation of log2(n) ---')
    approx_log2()

    print('\n--- Table of fixed Q16 dB ---')
    table_db_q16()

    print('')



import array
import lc3
import pytest


@pytest.mark.parametrize("bit_depth,decoded_length", [(16, 960), (24, 1440)])
def test_decode_with_bit_depth(bit_depth, decoded_length) -> None:
    decoder = lc3.Decoder(frame_duration_us=10000, sample_rate_hz=48000)
    decoded_frame = decoder.decode(bytes(120), bit_depth=bit_depth)
    assert isinstance(decoded_frame, bytes)
    assert len(decoded_frame) == decoded_length


def test_decode_without_bit_depth() -> None:
    decoder = lc3.Decoder(frame_duration_us=10000, sample_rate_hz=48000)
    decoded_frame = decoder.decode(bytes(120))
    assert isinstance(decoded_frame, array.array)
    assert len(decoded_frame) == 480
    assert all(isinstance(e, float) for e in decoded_frame)


def test_decode_with_bad_bit_depth() -> None:
    decoder = lc3.Decoder(frame_duration_us=10000, sample_rate_hz=48000)
    with pytest.raises(lc3.InvalidArgumentError):
        decoder.decode(bytes(120), bit_depth=128)


@pytest.mark.parametrize("bit_depth", [16, 24])
def test_encode_with_bit_depth(bit_depth) -> None:
    encoder = lc3.Encoder(frame_duration_us=10000, sample_rate_hz=48000)
    encoded_frame = encoder.encode(bytes(1920), num_bytes=120, bit_depth=bit_depth)
    assert isinstance(encoded_frame, bytes)
    assert len(encoded_frame) == 120


@pytest.mark.parametrize("pcm", [bytes(1920), [0.0] * 1920])
def test_encode_without_bit_depth(pcm) -> None:
    encoder = lc3.Encoder(frame_duration_us=10000, sample_rate_hz=48000)
    encoded_frame = encoder.encode(pcm, num_bytes=120, bit_depth=None)
    assert isinstance(encoded_frame, bytes)
    assert len(encoded_frame) == 120


def test_encode_with_bad_bit_depth() -> None:
    encoder = lc3.Encoder(frame_duration_us=10000, sample_rate_hz=48000)
    with pytest.raises(lc3.InvalidArgumentError):
        encoder.encode(bytes(1920), num_bytes=120, bit_depth=128)



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import scipy.signal as signal
import scipy.io.wavfile as wavfile
import struct
import argparse

import lc3
import tables as T, appendix_c as C

import mdct, energy, bwdet, sns, tns, spec, ltpf
import bitstream

class Decoder:

    def __init__(self, dt_ms, sr_hz):

        dt = { 7.5: T.DT_7M5, 10: T.DT_10M }[dt_ms]

        sr = {  8000: T.SRATE_8K , 16000: T.SRATE_16K, 24000: T.SRATE_24K,
               32000: T.SRATE_32K, 48000: T.SRATE_48K }[sr_hz]

        self.sr = sr
        self.ne = T.NE[dt][sr]
        self.ns = T.NS[dt][sr]

        self.mdct = mdct.MdctInverse(dt, sr)

        self.bwdet = bwdet.BandwidthDetector(dt, sr)
        self.spec = spec.SpectrumSynthesis(dt, sr)
        self.tns = tns.TnsSynthesis(dt)
        self.sns = sns.SnsSynthesis(dt, sr)
        self.ltpf = ltpf.LtpfSynthesis(dt, sr)

    def decode(self, data):

        b = bitstream.BitstreamReader(data)

        bw = self.bwdet.get(b)
        if bw > self.sr:
            raise ValueError('Invalid bandwidth indication')

        self.spec.load(b)

        self.tns.load(b, bw, len(data))

        pitch = b.read_bit()

        self.sns.load(b)

        if pitch:
            self.ltpf.load(b)
        else:
            self.ltpf.disable()

        x = self.spec.decode(b, bw, len(data))

        return (x, bw, pitch)

    def synthesize(self, x, bw, pitch, nbytes):

        x = self.tns.run(x, bw)

        x = self.sns.run(x)

        x = np.append(x, np.zeros(self.ns - self.ne))
        x = self.mdct.run(x)

        x = self.ltpf.run(x)

        return x

    def run(self, data):

        (x, bw, pitch) = self.decode(data)

        x = self.synthesize(x, bw, pitch, len(data))

        return x

def check_appendix_c(dt):

    i0 = dt - T.DT_7M5

    dec_c = lc3.setup_decoder(int(T.DT_MS[dt] * 1000), 16000)
    ok = True

    for i in range(len(C.BYTES_AC[i0])):

        pcm = lc3.decode(dec_c, bytes(C.BYTES_AC[i0][i]))
        ok = ok and np.max(np.abs(pcm - C.X_HAT_CLIP[i0][i])) < 1

    return ok

def check():

    ok = True

    for dt in range(T.DT_7M5, T.NUM_DT):
        ok = ok and check_appendix_c(dt)

    return ok



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import scipy.signal as signal
import scipy.io.wavfile as wavfile
import struct
import argparse

import lc3
import tables as T, appendix_c as C

import attdet, ltpf
import mdct, energy, bwdet, sns, tns, spec
import bitstream

class Encoder:

    def __init__(self, dt_ms, sr_hz):

        dt = { 7.5: T.DT_7M5, 10: T.DT_10M }[dt_ms]

        sr = {  8000: T.SRATE_8K , 16000: T.SRATE_16K, 24000: T.SRATE_24K,
               32000: T.SRATE_32K, 48000: T.SRATE_48K }[sr_hz]

        self.ne = T.NE[dt][sr]

        self.attdet = attdet.AttackDetector(dt, sr)
        self.ltpf = ltpf.LtpfAnalysis(dt, sr)

        self.mdct = mdct.MdctForward(dt, sr)
        self.energy = energy.EnergyBand(dt, sr)
        self.bwdet = bwdet.BandwidthDetector(dt, sr)
        self.sns = sns.SnsAnalysis(dt, sr)
        self.tns = tns.TnsAnalysis(dt)
        self.spec = spec.SpectrumAnalysis(dt, sr)

    def analyse(self, x, nbytes):

        att = self.attdet.run(nbytes, x)

        pitch_present = self.ltpf.run(x)

        x = self.mdct.run(x)[:self.ne]

        (e, nn_flag) = self.energy.compute(x)
        if nn_flag:
            self.ltpf.disable()

        bw = self.bwdet.run(e)

        x = self.sns.run(e, att, x)

        x = self.tns.run(x, bw, nn_flag, nbytes)

        (xq, lastnz, x) = self.spec.run(bw, nbytes,
            self.bwdet.get_nbits(), self.ltpf.get_nbits(),
            self.sns.get_nbits(), self.tns.get_nbits(), x)

        return pitch_present

    def encode(self, pitch_present, nbytes):

        b = bitstream.BitstreamWriter(nbytes)

        self.bwdet.store(b)

        self.spec.store(b)

        self.tns.store(b)

        b.write_bit(pitch_present)

        self.sns.store(b)

        if pitch_present:
            self.ltpf.store(b)

        self.spec.encode(b)

        return b.terminate()

    def run(self, x, nbytes):

        pitch_present = self.analyse(x, nbytes)

        data = self.encode(pitch_present, nbytes)

        return data

def check_appendix_c(dt):

    i0 = dt - T.DT_7M5

    enc_c = lc3.setup_encoder(int(T.DT_MS[dt] * 1000), 16000)
    ok = True

    for i in range(len(C.X_PCM[i0])):

        data = lc3.encode(enc_c, C.X_PCM[i0][i], C.NBYTES[i0])
        ok = ok and data == C.BYTES_AC[i0][i]

    return ok

def check():

    ok = True

    for dt in ( T.DT_7M5, T.DT_10M ):
        ok = ok and check_appendix_c(dt)

    return ok



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import scipy.signal as signal

import lc3
import tables as T, appendix_c as C

### ------------------------------------------------------------------------ ###

class Resampler_12k8:

    def __init__(self, dt, sr, history = 0):

        self.sr = sr
        self.p = 192 // T.SRATE_KHZ[sr]
        self.w = 240 // self.p

        self.n = ((T.DT_MS[dt] * 128) / 10).astype(int)
        self.d = [ 24, 44 ][dt == T.DT_7M5]

        self.x = np.zeros(self.w + T.NS[dt][sr])
        self.u = np.zeros(self.n + 2)
        self.y = np.zeros(self.n + self.d + history)

    def resample(self, x):

        p = self.p
        w = self.w
        d = self.d
        n = self.n

        ### Sliding window

        self.x[:w] = self.x[-w:]
        self.x[w:] = x
        self.u[:2] = self.u[-2:]

        if len(self.y) > 2*n + d:
            self.y[n+d:-n] = self.y[d+2*n:]
        if len(self.y) > n + d:
            self.y[-n:] = self.y[:n]
        self.y[:d] = self.y[n:d+n]

        x = self.x
        u = self.u

        ### Resampling

        h = np.zeros(240 + p)
        h[-119:] = T.LTPF_H12K8[:119]
        h[ :120] = T.LTPF_H12K8[119:]

        for i in range(n):
            e = (15 * i) // p
            f = (15 * i)  % p
            k = np.arange(-120, 120 + p, p) - f
            u[2+i] = p * np.dot( x[e:e+w+1], np.take(h, k) )

        if self.sr == T.SRATE_8K:
            u = 0.5 * u

        ### High-pass filtering

        b = [ 0.9827947082978771, -1.9655894165957540, 0.9827947082978771 ]
        a = [ 1                 , -1.9652933726226904, 0.9658854605688177 ]

        self.y[d:d+n] = b[0] * u[2:] + b[1] * u[1:-1] + b[2] * u[:-2]
        for i in range(n):
            self.y[d+i] -= a[1] * self.y[d+i-1] + a[2] * self.y[d+i-2]

        return self.y


class Resampler_6k4:

    def __init__(self, n, history = 0):

        self.x = np.zeros(n + 5)
        self.n = n // 2

        self.y = np.zeros(self.n + history)

    def resample(self, x):

        n = self.n

        ### Sliding window

        self.x[:3] = self.x[-5:-2]
        self.x[3:] = x[:2*n+2]
        x = self.x

        if len(self.y) > 2*n:
            self.y[n:-n] = self.y[2*n:]
        if len(self.y) > n:
            self.y[-n:] = self.y[:n]

        ### Downsampling to 6.4 KHz

        h = [ 0.1236796411180537, 0.2353512128364889, 0.2819382920909148,
              0.2353512128364889, 0.1236796411180537 ]

        self.y[:n] = [ np.dot(x[2*i:2*i+5], h) for i in range(self.n) ]
        return self.y


def initial_hp50_state():
    return { 's1': 0, 's2': 0 }

### ------------------------------------------------------------------------ ###

class Ltpf:

    def __init__(self, dt, sr):

        self.dt = dt
        self.sr = sr

        (self.pitch_present, self.pitch_index) = (None, None)


class LtpfAnalysis(Ltpf):

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

        self.resampler_12k8 = Resampler_12k8(dt, sr,
            history = 232 + (32 if dt == T.DT_2M5 else 0))

        self.resampler_6k4 = Resampler_6k4(self.resampler_12k8.n,
            history = 114 + (16 if dt == T.DT_2M5 else 0))

        self.active = False
        self.tc = 0
        self.pitch = 0
        self.nc = np.zeros(2)

    def get_data(self):

        return { 'active' : self.active,
                 'pitch_index' : self.pitch_index }

    def get_nbits(self):

        return 1 + 10 * int(self.pitch_present)

    def correlate(self, x, i0, n, k0, k1):

        return np.array([ np.dot(
            np.take(x, np.arange(i0, n)),
            np.take(x, np.arange(i0, n) - k)) for k in range(k0, 1+k1) ])

    def norm_corr(self, x, i0, n, k):

        u  = np.take(x, np.arange(i0, n))
        v  = np.take(x, np.arange(i0, n) - k)
        uv = np.dot(u, v)
        return uv / np.sqrt(np.dot(u, u) * np.dot(v, v)) if uv > 0 else 0

    def run(self, x):

        ### Resampling

        x_12k8 = self.resampler_12k8.resample(x)

        ### Pitch detection algorithm

        x  = self.resampler_6k4.resample(x_12k8)
        i0 = [-16, 0][self.dt > T.DT_2M5]
        n  = self.resampler_6k4.n

        r  = self.correlate(x, i0, n, 17, 114)
        rw = r * (1 - 0.5 * np.arange(len(r)) / (len(r) - 1))

        tc = self.tc
        k0 = max(0, tc-4)
        k1 = min(len(r)-1, tc+4)
        t  = [ 17 + np.argmax(rw), 17 + k0 + np.argmax(r[k0:1+k1]) ]

        nc = [ self.norm_corr(x, i0, n, t[i]) for i in range(2) ]
        ti = int(nc[1] > 0.85 * nc[0])
        self.tc = t[ti] - 17

        self.pitch_present = bool(nc[ti] > 0.6)

        ### Pitch-lag parameter

        if self.pitch_present:
            tc = self.tc + 17

            x  = x_12k8
            i0 = [-32, 0][self.dt > T.DT_2M5]
            n  = self.resampler_12k8.n

            k0 = max( 32, 2*tc-4)
            k1 = min(228, 2*tc+4)
            r  = self.correlate(x, i0, n, k0-4, k1+4)
            e  = k0 + np.argmax(r[4:-4])

            h = np.zeros(42)
            h[-15:] = T.LTPF_H4[:15]
            h[ :16] = T.LTPF_H4[15:]

            m = np.arange(-4, 5)
            s = [ np.dot( np.take(r, e-k0+4 + m), np.take(h, 4*m-d) ) \
                      for d in range(-3, 4) ]

            f = np.argmax(s[3:])            if e <=  32 else \
                -3 + np.argmax(s)           if e <  127 else \
                -2 + 2*np.argmax(s[1:-1:2]) if e <  157 else 0

            e -=   (f < 0)
            f += 4*(f < 0)

            self.pitch_index = 4*e + f    - 128 if e < 127 else \
                               2*e + f//2 + 126 if e < 157 else e + 283

        else:
            e = f = 0
            self.pitch_index = 0

        ### Activation bit

        h = np.zeros(24)
        h[-7:] = T.LTPF_HI[:7]
        h[ :8] = T.LTPF_HI[7:]

        x  = x_12k8
        i0 = [-32, 0][self.dt > T.DT_2M5]
        n  = self.resampler_12k8.n

        k = np.arange(-2, 3)
        u = [ np.dot( np.take(x, i-k), np.take(h, 4*k) ) \
                  for i in range(i0, n) ]
        v = [ np.dot( np.take(x, i-k), np.take(h, 4*k-f) ) \
                  for i in range(i0-e, n-e) ]

        nc = max(0, np.dot(u, v)) / np.sqrt(np.dot(u, u) * np.dot(v, v)) \
                if self.pitch_present else 0

        pitch = e + f/4

        if not self.active:
            active = (self.dt == T.DT_10M or self.nc[1] > 0.94) \
                     and self.nc[0] > 0.94 and nc > 0.94

        else:
            dp = abs(pitch - self.pitch)
            dc = nc - self.nc[0]
            active = nc > 0.9 or (dp < 2 and dc > -0.1 and nc > 0.84)

        if not self.pitch_present:
            active = False
            pitch = 0
            nc = 0

        self.active = active
        self.pitch = pitch
        self.nc[1] = self.nc[0]
        self.nc[0] = nc

        return self.pitch_present

    def disable(self):

        self.active = False

    def store(self, b):

        b.write_uint(self.active, 1)
        b.write_uint(self.pitch_index, 9)


class LtpfSynthesis(Ltpf):

    C_N = [ T.LTPF_N_8K , T.LTPF_N_16K,
            T.LTPF_N_24K, T.LTPF_N_32K, T.LTPF_N_48K ]

    C_D = [ T.LTPF_D_8K , T.LTPF_D_16K,
            T.LTPF_D_24K, T.LTPF_D_32K, T.LTPF_D_48K ]

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

        self.C_N = LtpfSynthesis.C_N[sr]
        self.C_D = LtpfSynthesis.C_D[sr]

        ns = T.NS[dt][sr]

        self.active = [ False, False ]
        self.pitch_index = 0

        max_pitch_12k8 = 228
        max_pitch = max_pitch_12k8 * T.SRATE_KHZ[self.sr] / 12.8
        max_pitch = np.ceil(max_pitch).astype(int)

        self.x = np.zeros(ns)
        self.y = np.zeros(max_pitch + len(self.C_D[0]))

        self.p_e = [ 0, 0 ]
        self.p_f = [ 0, 0 ]
        self.c_n = [ None, None ]
        self.c_d = [ None, None ]

    def load(self, b):

        self.active[0] = bool(b.read_uint(1))
        self.pitch_index = b.read_uint(9)

    def disable(self):

        self.active[0] = False
        self.pitch_index = 0

    def run(self, x, nbytes):

        sr = self.sr
        dt = self.dt

        ### Filter parameters

        pitch_index = self.pitch_index

        if pitch_index >= 440:
            p_e = pitch_index - 283
            p_f = 0
        elif pitch_index >= 380:
            p_e = pitch_index // 2 - 63
            p_f = 2*(pitch_index - 2*(p_e + 63))
        else:
            p_e = pitch_index // 4 + 32
            p_f = pitch_index - 4*(p_e - 32)

        p = (p_e + p_f / 4) * T.SRATE_KHZ[self.sr] / 12.8

        self.p_e[0] = int(p * 4 + 0.5) // 4
        self.p_f[0] = int(p * 4 + 0.5) - 4*self.p_e[0]

        nbits = round(nbytes*8 * 10 / T.DT_MS[dt])
        if dt == T.DT_2M5:
            nbits = int(nbits * (1 - 0.4))
        elif dt == T.DT_5M:
            nbits = nbits - 160

        g_idx = max(nbits // 80, 3+sr) - (3+sr)

        g = [ 0.4, 0.35, 0.3, 0.25 ][g_idx] if g_idx < 4 else 0
        g_idx = min(g_idx, 3)

        self.c_n[0] = 0.85 * g * LtpfSynthesis.C_N[sr][g_idx]
        self.c_d[0] = g * LtpfSynthesis.C_D[sr][self.p_f[0]]

        ### Transition handling

        n0 = (T.SRATE_KHZ[sr] * 1000) // 400
        ns = T.NS[dt][sr]

        x  = np.append(x, self.x)
        y  = np.append(np.zeros(ns), self.y)
        yc = y.copy()

        c_n = self.c_n
        c_d = self.c_d

        l_n = len(c_n[0])
        l_d = len(c_d[0])

        d = [ self.p_e[0] - (l_d - 1) // 2,
              self.p_e[1] - (l_d - 1) // 2 ]

        for k in range(n0):

            if not self.active[0] and not self.active[1]:
                y[k] = x[k]

            elif self.active[0] and not self.active[1]:
                u = np.dot(c_n[0], np.take(x, k - np.arange(l_n))) - \
                    np.dot(c_d[0], np.take(y, k - d[0] - np.arange(l_d)))
                y[k] = x[k] - (k/n0) * u

            elif not self.active[0] and self.active[1]:
                u = np.dot(c_n[1], np.take(x, k - np.arange(l_n))) - \
                    np.dot(c_d[1], np.take(y, k - d[1] - np.arange(l_d)))
                y[k] = x[k] - (1 - k/n0) * u

            elif self.p_e[0] == self.p_e[1] and self.p_f[0] == self.p_f[1]:
                u = np.dot(c_n[0], np.take(x, k - np.arange(l_n))) - \
                    np.dot(c_d[0], np.take(y, k - d[0] - np.arange(l_d)))
                y[k] = x[k] - u

            else:
                u = np.dot(c_n[1], np.take(x, k - np.arange(l_n))) - \
                    np.dot(c_d[1], np.take(y, k - d[1] - np.arange(l_d)))
                yc[k] = x[k] - (1 - k/n0) * u

                u = np.dot(c_n[0], np.take(yc, k - np.arange(l_n))) - \
                    np.dot(c_d[0], np.take(y , k - d[0] - np.arange(l_d)))
                y[k] = yc[k] - (k/n0) * u

        ### Remainder of the frame

        for k in range(n0, ns):

            if not self.active[0]:
                y[k] = x[k]

            else:
                u = np.dot(c_n[0], np.take(x, k - np.arange(l_n))) - \
                    np.dot(c_d[0], np.take(y, k - d[0] - np.arange(l_d)))
                y[k] = x[k] - u

        ### Sliding window

        self.active[1] = self.active[0]
        self.p_e[1] = self.p_e[0]
        self.p_f[1] = self.p_f[0]
        self.c_n[1] = self.c_n[0]
        self.c_d[1] = self.c_d[0]

        self.x = x[:ns]
        self.y = np.append(self.y[ns:], y[:ns])

        return y[:ns]

def initial_state():
    return { 'active' : False, 'pitch': 0, 'nc':  np.zeros(2),
             'hp50' : initial_hp50_state(),
             'x_12k8' : np.zeros(384), 'x_6k4' : np.zeros(178), 'tc' : 0 }

def initial_sstate():
    return { 'active': False, 'pitch': 0,
             'c': np.zeros(2*12), 'x': np.zeros(12) }

### ------------------------------------------------------------------------ ###

def check_resampler(rng, dt, sr):

    ns = T.NS[dt][sr]
    nt = (5 * T.SRATE_KHZ[sr]) // 4
    ok = True

    r = Resampler_12k8(dt, sr)

    hp50_c = initial_hp50_state()
    x_c = np.zeros(nt)
    y_c = np.zeros(384)

    for run in range(10):

        x = ((2 * rng.random(ns)) - 1) * (2 ** 15 - 1)
        y = r.resample(x)

        x_c = np.append(x_c[-nt:], x.astype(np.int16))
        y_c[:-r.n] = y_c[r.n:]
        y_c = lc3.ltpf_resample(dt, sr, hp50_c, x_c, y_c)

        ok = ok and np.amax(np.abs(y_c[-r.d-r.n:] - y[:r.d+r.n]/2)) < 4

    return ok

def check_resampler_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    nt = (5 * T.SRATE_KHZ[sr]) // 4
    n  = [ 96, 128 ][i0]
    k  = [ 44,  24 ][i0] + n

    state = initial_hp50_state()

    x = np.append(np.zeros(nt), C.X_PCM[i0][0])
    y = np.zeros(384)
    y = lc3.ltpf_resample(dt, sr, state, x, y)
    u = y[-k:len(C.X_TILDE_12K8D[i0][0])-k]

    ok = ok and np.amax(np.abs(u - C.X_TILDE_12K8D[i0][0]/2)) < 2

    x = np.append(x[-nt:], C.X_PCM[i0][1])
    y[:-n] = y[n:]
    y = lc3.ltpf_resample(dt, sr, state, x, y)
    u = y[-k:len(C.X_TILDE_12K8D[i0][1])-k]

    ok = ok and np.amax(np.abs(u - C.X_TILDE_12K8D[i0][1]/2)) < 2

    return ok

def check_analysis(rng, dt, sr):

    ns = T.NS[dt][sr]
    nt = (5 * T.SRATE_KHZ[sr]) // 4
    ok = True

    state_c = initial_state()
    x_c = np.zeros(ns+nt)

    ltpf = LtpfAnalysis(dt, sr)

    t = np.arange(100 * ns) / (T.SRATE_KHZ[sr] * 1000)
    s = signal.chirp(t, f0=10, f1=2500, t1=t[-1], method='logarithmic')

    for i in range(20):

        x = s[i*ns:(i+1)*ns] * (2 ** 15 - 1)

        pitch_present = ltpf.run(x)
        data = ltpf.get_data()

        x_c = np.append(x_c[-nt:], x.astype(np.int16))
        (pitch_present_c, data_c) = lc3.ltpf_analyse(dt, sr, state_c, x_c)

        ok = ok and (not pitch_present or state_c['tc'] == ltpf.tc)
        ok = ok and np.amax(np.abs(state_c['nc'][0] - ltpf.nc[0])) < 1e-1
        ok = ok and pitch_present_c == pitch_present
        ok = ok and data_c['active'] == data['active']
        ok = ok and data_c['pitch_index'] == data['pitch_index']
        ok = ok and lc3.ltpf_get_nbits(pitch_present) == ltpf.get_nbits()

    return ok

def check_synthesis(rng, dt, sr):

    ok = True

    ns = T.NS[dt][sr]
    nd = 18 * T.SRATE_KHZ[sr]

    synthesis = LtpfSynthesis(dt, sr)

    state_c = initial_sstate()
    x_c = np.zeros(nd+ns)

    for i in range(50):

        pitch_present = bool(rng.integers(0, 10) >= 1)
        if not pitch_present:
            synthesis.disable()
        else:
            synthesis.active[0] = bool(rng.integers(0, 5) >= 1)
            synthesis.pitch_index = rng.integers(0, 512)

        data_c = None if not pitch_present else \
            { 'active' : synthesis.active[0],
              'pitch_index' : synthesis.pitch_index }

        x = rng.random(ns) * 1e4
        nbytes = rng.integers(10*(2+sr), 10*(6+sr))

        x_c[:nd] = x_c[ns:]
        x_c[nd:] = x

        y = synthesis.run(x, nbytes)
        x_c = lc3.ltpf_synthesize(dt, sr, nbytes, state_c, data_c, x_c)

        ok = ok and np.amax(np.abs(x_c[nd:] - y)) < 1e-2

    return ok

def check_analysis_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    nt = (5 * T.SRATE_KHZ[sr]) // 4

    state = initial_state()

    x = np.append(np.zeros(nt), C.X_PCM[i0][0])
    (pitch_present, data) = lc3.ltpf_analyse(dt, sr, state, x)

    ok = ok and C.T_CURR[i0][0] - state['tc'] == 17
    ok = ok and np.amax(np.abs(state['nc'][0] - C.NC_LTPF[i0][0])) < 1e-5
    ok = ok and pitch_present == C.PITCH_PRESENT[i0][0]
    ok = ok and data['pitch_index'] == C.PITCH_INDEX[i0][0]
    ok = ok and data['active'] == C.LTPF_ACTIVE[i0][0]

    x = np.append(x[-nt:], C.X_PCM[i0][1])
    (pitch_present, data) = lc3.ltpf_analyse(dt, sr, state, x)

    ok = ok and C.T_CURR[i0][1] - state['tc'] == 17
    ok = ok and np.amax(np.abs(state['nc'][0] - C.NC_LTPF[i0][1])) < 1e-5
    ok = ok and pitch_present == C.PITCH_PRESENT[i0][1]
    ok = ok and data['pitch_index'] == C.PITCH_INDEX[i0][1]
    ok = ok and data['active'] == C.LTPF_ACTIVE[i0][1]

    return ok

def check_synthesis_appendix_c(dt):

    sr = T.SRATE_16K

    ok = True
    if dt != T.DT_10M:
        return ok

    ns = T.NS[dt][sr]
    nd = 18 * T.SRATE_KHZ[sr]

    NBYTES = [ C.LTPF_C2_NBITS // 8, C.LTPF_C3_NBITS // 8,
               C.LTPF_C4_NBITS // 8, C.LTPF_C5_NBITS // 8 ]

    ACTIVE = [ C.LTPF_C2_ACTIVE, C.LTPF_C3_ACTIVE,
               C.LTPF_C4_ACTIVE, C.LTPF_C5_ACTIVE ]

    PITCH_INDEX = [ C.LTPF_C2_PITCH_INDEX, C.LTPF_C3_PITCH_INDEX,
                    C.LTPF_C4_PITCH_INDEX, C.LTPF_C5_PITCH_INDEX ]

    X = [ C.LTPF_C2_X, C.LTPF_C3_X,
          C.LTPF_C4_X, C.LTPF_C5_X ]

    PREV = [ C.LTPF_C2_PREV, C.LTPF_C3_PREV,
             C.LTPF_C4_PREV, C.LTPF_C5_PREV  ]

    TRANS = [ C.LTPF_C2_TRANS, C.LTPF_C3_TRANS,
              C.LTPF_C4_TRANS, C.LTPF_C5_TRANS ]

    for i in range(4):

        state = initial_sstate()
        nbytes = NBYTES[i]

        data = { 'active' : ACTIVE[i][0], 'pitch_index' : PITCH_INDEX[i][0] }
        x = np.append(np.zeros(nd), X[i][0])

        lc3.ltpf_synthesize(dt, sr, nbytes, state, data, x)

        data = { 'active' : ACTIVE[i][1], 'pitch_index' : PITCH_INDEX[i][1] }
        x[  :nd-ns] = PREV[i][0][-nd+ns:]
        x[nd-ns:nd] = PREV[i][1]
        x[nd:nd+ns] = X[i][1]

        y = lc3.ltpf_synthesize(dt, sr, nbytes, state, data, x)[nd:]

        ok = ok and np.amax(np.abs(y - TRANS[i])) < 1e-3

    return ok

def check():

    rng = np.random.default_rng(1234)
    ok = True

    for dt in range(T.NUM_DT):
        for sr in range(T.SRATE_8K, T.SRATE_48K + 1):
            ok = ok and check_resampler(rng, dt, sr)
            ok = ok and check_analysis(rng, dt, sr)
            ok = ok and check_synthesis(rng, dt, sr)

    for dt in ( T.DT_7M5, T.DT_10M ):
        ok = ok and check_resampler_appendix_c(dt)
        ok = ok and check_analysis_appendix_c(dt)
        ok = ok and check_synthesis_appendix_c(dt)

    return ok

### ------------------------------------------------------------------------ ###



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

import lc3
import tables as T, appendix_c as C

### ------------------------------------------------------------------------ ###

class Tns:
    SUB_LIM_2M5_NB   = [ [   3,  10,  20  ] ]
    SUB_LIM_2M5_WB   = [ [   3,  20,  40  ] ]
    SUB_LIM_2M5_SSWB = [ [   3,  30,  60  ] ]
    SUB_LIM_2M5_SWB  = [ [   3,  40,  80  ] ]
    SUB_LIM_2M5_FB   = [ [   3,  51, 100  ] ]

    SUB_LIM_2M5 = [
        SUB_LIM_2M5_NB , SUB_LIM_2M5_WB, SUB_LIM_2M5_SSWB,
        SUB_LIM_2M5_SWB, SUB_LIM_2M5_FB, SUB_LIM_2M5_FB, SUB_LIM_2M5_FB ]

    SUB_LIM_5M_NB    = [ [  6,  23,  40 ] ]
    SUB_LIM_5M_WB    = [ [  6,  43,  80 ] ]
    SUB_LIM_5M_SSWB  = [ [  6,  63, 120 ] ]
    SUB_LIM_5M_SWB   = [ [  6,  43,  80 ], [  80, 120, 160 ] ]
    SUB_LIM_5M_FB    = [ [  6,  53, 100 ], [ 100, 150, 200 ] ]

    SUB_LIM_5M = [
        SUB_LIM_5M_NB , SUB_LIM_5M_WB, SUB_LIM_5M_SSWB,
        SUB_LIM_5M_SWB, SUB_LIM_5M_FB, SUB_LIM_5M_FB, SUB_LIM_5M_FB ]

    SUB_LIM_7M5_NB   = [ [   9,  26,  43,  60 ] ]
    SUB_LIM_7M5_WB   = [ [   9,  46,  83, 120 ] ]
    SUB_LIM_7M5_SSWB = [ [   9,  66, 123, 180 ] ]
    SUB_LIM_7M5_SWB  = [ [   9,  46,  82, 120 ], [ 120, 159, 200, 240 ] ]
    SUB_LIM_7M5_FB   = [ [   9,  56, 103, 150 ], [ 150, 200, 250, 300 ] ]

    SUB_LIM_7M5 = [
        SUB_LIM_7M5_NB , SUB_LIM_7M5_WB, SUB_LIM_7M5_SSWB,
        SUB_LIM_7M5_SWB, SUB_LIM_7M5_FB, None, None ]

    SUB_LIM_10M_NB   = [ [  12,  34,  57,  80 ] ]
    SUB_LIM_10M_WB   = [ [  12,  61, 110, 160 ] ]
    SUB_LIM_10M_SSWB = [ [  12,  88, 164, 240 ] ]
    SUB_LIM_10M_SWB  = [ [  12,  61, 110, 160 ], [ 160, 213, 266, 320 ] ]
    SUB_LIM_10M_FB   = [ [  12,  74, 137, 200 ], [ 200, 266, 333, 400 ] ]

    SUB_LIM_10M = [
        SUB_LIM_10M_NB , SUB_LIM_10M_WB, SUB_LIM_10M_SSWB,
        SUB_LIM_10M_SWB, SUB_LIM_10M_FB, SUB_LIM_10M_FB, SUB_LIM_10M_FB ]

    SUB_LIM = [ SUB_LIM_2M5, SUB_LIM_5M, SUB_LIM_7M5, SUB_LIM_10M ]


    FREQ_LIM_2M5_NB   = [   3,  20 ]
    FREQ_LIM_2M5_WB   = [   3,  40 ]
    FREQ_LIM_2M5_SSWB = [   3,  60 ]
    FREQ_LIM_2M5_SWB  = [   3,  80 ]
    FREQ_LIM_2M5_FB   = [   3, 100 ]

    FREQ_LIM_2M5 = [
        FREQ_LIM_2M5_NB , FREQ_LIM_2M5_WB, FREQ_LIM_2M5_SSWB,
        FREQ_LIM_2M5_SWB, FREQ_LIM_2M5_FB, FREQ_LIM_2M5_FB, FREQ_LIM_2M5_FB ]

    FREQ_LIM_5M_NB    = [   6,  40 ]
    FREQ_LIM_5M_WB    = [   6,  80 ]
    FREQ_LIM_5M_SSWB  = [   6, 120 ]
    FREQ_LIM_5M_SWB   = [   6,  80, 160 ]
    FREQ_LIM_5M_FB    = [   6, 100, 200 ]

    FREQ_LIM_5M = [
        FREQ_LIM_5M_NB , FREQ_LIM_5M_WB, FREQ_LIM_5M_SSWB,
        FREQ_LIM_5M_SWB, FREQ_LIM_5M_FB, FREQ_LIM_5M_FB, FREQ_LIM_5M_FB ]

    FREQ_LIM_7M5_NB   = [   9,  60 ]
    FREQ_LIM_7M5_WB   = [   9, 120 ]
    FREQ_LIM_7M5_SSWB = [   9, 180 ]
    FREQ_LIM_7M5_SWB  = [   9, 120, 240 ]
    FREQ_LIM_7M5_FB   = [   9, 150, 300 ]

    FREQ_LIM_7M5 = [
        FREQ_LIM_7M5_NB , FREQ_LIM_7M5_WB, FREQ_LIM_7M5_SSWB,
        FREQ_LIM_7M5_SWB, FREQ_LIM_7M5_FB, None, None ]

    FREQ_LIM_10M_NB   = [  12,  80 ]
    FREQ_LIM_10M_WB   = [  12, 160 ]
    FREQ_LIM_10M_SSWB = [  12, 240 ]
    FREQ_LIM_10M_SWB  = [  12, 160, 320 ]
    FREQ_LIM_10M_FB   = [  12, 200, 400 ]

    FREQ_LIM_10M = [
        FREQ_LIM_10M_NB , FREQ_LIM_10M_WB, FREQ_LIM_10M_SSWB,
        FREQ_LIM_10M_SWB, FREQ_LIM_10M_FB, FREQ_LIM_10M_FB, FREQ_LIM_10M_FB ]

    FREQ_LIM = [ FREQ_LIM_2M5, FREQ_LIM_5M, FREQ_LIM_7M5, FREQ_LIM_10M ]


    def __init__(self, dt):

        self.dt = dt

        (self.nfilters, self.lpc_weighting, self.rc_order, self.rc) = \
            (0, False, np.array([ 0, 0 ]), np.array([ 0, 0 ]))

    def get_data(self):

        rc = np.append(self.rc - 8, np.zeros((2, 8 - len(self.rc[0]))), axis=1)

        return { 'nfilters' : self.nfilters,
                 'lpc_weighting' : self.lpc_weighting,
                 'rc_order' : self.rc_order, 'rc' : rc }

    def get_nbits(self):

        lpc_weighting = self.lpc_weighting
        nbits = 0

        for f in range(self.nfilters):
            rc_order = self.rc_order[f]
            rc = self.rc[f]

            nbits_order = T.TNS_ORDER_BITS[int(lpc_weighting)][rc_order]
            nbits_coef = sum([ T.TNS_COEF_BITS[k][rc[k]]
                                  for k in range(rc_order) ])

            nbits += ((2048 + nbits_order + nbits_coef) + 2047) >> 11

        return nbits


class TnsAnalysis(Tns):

    def __init__(self, dt):

        super().__init__(dt)

    def compute_lpc_coeffs(self, bw, f, x):

        ### Normalized autocorrelation function

        S = Tns.SUB_LIM[self.dt][bw][f]
        maxorder = [ 4, 8 ][self.dt > T.DT_5M]

        r = np.append([ 3 ], np.zeros(maxorder))
        e = [ sum(x[S[s]:S[s+1]] ** 2) for s in range(len(S)-1) ]

        for k in range(len(r) if sum(e) > 0 else 0):
            c = [ np.dot(x[S[s]:S[s+1]-k], x[S[s]+k:S[s+1]])
                      for s in range(len(S)-1) ]

            r[k] = np.sum( np.array(c) / np.array(e) )

        r *= np.exp(-0.5 * (0.02 * np.pi * np.arange(1+maxorder)) ** 2)

        ### Levinson-Durbin recursion

        err = r[0]
        a = np.ones(len(r))

        for k in range(1, len(a)):

            rc = -sum(a[:k] * r[k:0:-1]) / err

            a[1:k] += rc * a[k-1:0:-1]
            a[k] = rc

            err *= 1 - rc ** 2

        return (r[0] / err, a)

    def lpc_weight(self, pred_gain, a):

        gamma = 1 - (1 - 0.85) * (2 - pred_gain) / (2 - 1.5)
        return a * np.power(gamma, np.arange(len(a)))

    def coeffs_reflexion(self, a):

        rc = np.zeros(len(a)-1)
        b  = a.copy()

        for k in range(len(rc), 0, -1):
            rc[k-1] = b[k]
            e = 1 - rc[k-1] ** 2
            b[1:k] = (b[1:k] - rc[k-1] * b[k-1:0:-1]) / e

        return rc

    def quantization(self, rc, lpc_weighting):

        delta = np.pi / 17
        rc_i = np.rint(np.arcsin(rc) / delta).astype(int) + 8
        rc_q = np.sin(delta * (rc_i - 8))
        rc_q = np.rint(rc_q * 2**15) / 2**15

        rc_order = len(rc_i) - np.argmin(rc_i[::-1] == 8)

        return (rc_order, rc_q, rc_i)

    def filtering(self, st, x, rc_order, rc):

        y = np.empty(len(x))

        for i in range(len(x)):

            xi = x[i]
            s1 = xi

            for k in range(rc_order):
                s0 = st[k]
                st[k] = s1

                s1  = rc[k] * xi + s0
                xi += rc[k] * s0

            y[i] = xi

        return y

    def run(self, x, bw, nn_flag, nbytes):

        fstate = np.zeros(8)
        y = x.copy()

        self.nfilters = len(Tns.SUB_LIM[self.dt][bw])
        maxorder = [ 4, 8 ][self.dt > T.DT_5M]

        self.lpc_weighting = nbytes < 120 * (1 + self.dt) / 8

        self.rc_order = np.zeros(2, dtype=np.intc)
        self.rc = np.zeros((2, maxorder), dtype=np.intc)

        for f in range(self.nfilters):

            (pred_gain, a) = self.compute_lpc_coeffs(bw, f, x)

            tns_off = pred_gain <= 1.5 or nn_flag
            if tns_off:
                continue

            if self.lpc_weighting and pred_gain < 2:
                a = self.lpc_weight(pred_gain, a)

            rc = self.coeffs_reflexion(a)

            (rc_order, rc_q, rc_i) = \
                self.quantization(rc, self.lpc_weighting)

            self.rc_order[f] = rc_order
            self.rc[f] = rc_i

            if rc_order > 0:
                i0 = Tns.FREQ_LIM[self.dt][bw][f]
                i1 = Tns.FREQ_LIM[self.dt][bw][f+1]

                y[i0:i1] = self.filtering(
                    fstate, x[i0:i1], rc_order, rc_q)

        return y

    def store(self, b):

        for f in range(self.nfilters):
            lpc_weighting = self.lpc_weighting
            rc_order = self.rc_order[f]
            rc = self.rc[f]

            b.write_bit(min(rc_order, 1))

            if rc_order > 0:
                b.ac_encode(
                    T.TNS_ORDER_CUMFREQ[int(lpc_weighting)][rc_order-1],
                    T.TNS_ORDER_FREQ[int(lpc_weighting)][rc_order-1]    )

            for k in range(rc_order):
                b.ac_encode(T.TNS_COEF_CUMFREQ[k][rc[k]],
                            T.TNS_COEF_FREQ[k][rc[k]]    )


class TnsSynthesis(Tns):

    def filtering(self, st, x, rc_order, rc):

        y = x.copy()

        for i in range(len(x)):

            xi = x[i] - rc[rc_order-1] * st[rc_order-1]
            for k in range(rc_order-2, -1, -1):
                xi -= rc[k] * st[k]
                st[k+1] = xi * rc[k] + st[k]
            st[0] = xi

            y[i] = xi

        return y

    def load(self, b, bw, nbytes):

        self.nfilters = len(Tns.SUB_LIM[self.dt][bw])
        self.lpc_weighting = nbytes < 120 * (1 + self.dt) / 8
        self.rc_order = np.zeros(2, dtype=np.intc)
        self.rc = 8 * np.ones((2, 8), dtype=np.intc)

        for f in range(self.nfilters):

            if not b.read_bit():
                continue

            rc_order = 1 + b.ac_decode(
                T.TNS_ORDER_CUMFREQ[int(self.lpc_weighting)],
                T.TNS_ORDER_FREQ[int(self.lpc_weighting)])

            self.rc_order[f] = rc_order

            for k in range(rc_order):
                rc = b.ac_decode(T.TNS_COEF_CUMFREQ[k], T.TNS_COEF_FREQ[k])
                self.rc[f][k] = rc

    def run(self, x, bw):

        fstate = np.zeros(8)
        y = x.copy()

        for f in range(self.nfilters):

            rc_order = self.rc_order[f]
            rc = np.sin((np.pi / 17) * (self.rc[f] - 8))
            rc = np.rint(rc * 2**15) / 2**15

            if rc_order > 0:
                i0 = Tns.FREQ_LIM[self.dt][bw][f]
                i1 = Tns.FREQ_LIM[self.dt][bw][f+1]

                y[i0:i1] = self.filtering(
                    fstate, x[i0:i1], rc_order, rc)

        return y


### ------------------------------------------------------------------------ ###

def check_analysis(rng, dt, bw):

    ok = True

    analysis = TnsAnalysis(dt)
    nbytes_lim = int((48 * T.DT_MS[dt]) // 8)

    for i in range(10):
        ne = T.I[dt][bw][-1]
        x  = rng.random(ne) * 1e2
        x  = pow(x, .5 + i/5)

        for nn_flag in (True, False):
            for nbytes in (nbytes_lim, nbytes_lim + 1):

                y = analysis.run(x, bw, nn_flag, nbytes)
                (y_c, data_c) = lc3.tns_analyze(dt, bw, nn_flag, nbytes, x)

                ok = ok and data_c['lpc_weighting'] == analysis.lpc_weighting
                ok = ok and data_c['nfilters'] == analysis.nfilters
                for f in range(analysis.nfilters):
                    rc_order = analysis.rc_order[f]
                    rc_order_c = data_c['rc_order'][f]
                    rc_c = 8 + data_c['rc'][f]
                    ok = ok and rc_order_c == rc_order
                    ok = ok and not np.any(rc_c[:rc_order] - analysis.rc[f][:rc_order])

                ok = ok and lc3.tns_get_nbits(data_c) == analysis.get_nbits()
                ok = ok and np.amax(np.abs(y_c - y)) < 1e-2

    return ok

def check_synthesis(rng, dt, bw):

    ok = True
    synthesis = TnsSynthesis(dt)

    for i in range(100):

        ne = T.I[dt][bw][-1]
        x  = rng.random(ne) * 1e2

        maxorder = [ 4, 8 ][dt > T.DT_5M]
        synthesis.nfilters = 1 + int(dt >= T.DT_5M and bw >= T.SRATE_32K)
        synthesis.rc_order = rng.integers(0, 1+maxorder, 2)
        synthesis.rc = rng.integers(0, 17, 16).reshape(2, 8)

        y = synthesis.run(x, bw)
        y_c = lc3.tns_synthesize(dt, bw, synthesis.get_data(), x)

        ok = ok and np.amax(np.abs(y_c - y) < 1e-4)

    return ok

def check_analysis_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    fs = Tns.FREQ_LIM[i0][sr][0]
    fe = Tns.FREQ_LIM[i0][sr][1]
    st = np.zeros(8)

    for i in range(len(C.X_S[i0])):

        (_, a) = lc3.tns_compute_lpc_coeffs(dt, sr, C.X_S[i0][i])
        ok = ok and np.amax(np.abs(a[0] - C.TNS_LEV_A[i0][i])) < 1e-5

        rc = lc3.tns_lpc_reflection(dt, a[0])
        ok = ok and np.amax(np.abs(rc - C.TNS_LEV_RC[i0][i])) < 1e-5

        (rc_order, rc_i) = lc3.tns_quantize_rc(dt, C.TNS_LEV_RC[i0][i])
        ok = ok and rc_order == C.RC_ORDER[i0][i][0]
        ok = ok and np.any((rc_i + 8) - C.RC_I_1[i0][i] == 0)

        rc_q = lc3.tns_unquantize_rc(rc_i, rc_order)
        ok = ok and np.amax(np.abs(rc_q - C.RC_Q_1[i0][i])) < 1e-6

        (x, side) = lc3.tns_analyze(dt, sr, False, C.NBYTES[i0], C.X_S[i0][i])
        ok = ok and side['nfilters'] == 1
        ok = ok and side['rc_order'][0] == C.RC_ORDER[i0][i][0]
        ok = ok and not np.any((side['rc'][0] + 8) - C.RC_I_1[i0][i])
        ok = ok and lc3.tns_get_nbits(side) == C.NBITS_TNS[i0][i]
        ok = ok and np.amax(np.abs(x - C.X_F[i0][i])) < 1e-3

    return ok

def check_synthesis_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    for i in range(len(C.X_HAT_Q[i0])):

        side = {
            'nfilters' : 1,
            'lpc_weighting' : C.NBYTES[i0] < 120 * (1 + dt) / 8,
            'rc_order': C.RC_ORDER[i0][i],
            'rc': [ C.RC_I_1[i0][i] - 8, C.RC_I_2[i0][i] - 8 ]
        }

        g_int = C.GG_IND_ADJ[i0][i] + C.GG_OFF[i0][i]
        x = C.X_HAT_Q[i0][i] * (10 ** (g_int / 28))

        x = lc3.tns_synthesize(dt, sr, side, x)
        ok = ok and np.amax(np.abs(x - C.X_HAT_TNS[i0][i])) < 1e-3

    sr = T.SRATE_48K
    if dt != T.DT_10M:
        return ok

    side = {
        'nfilters' : 2,
        'lpc_weighting' : False,
        'rc_order': C.RC_ORDER_48K_10M,
        'rc': [ C.RC_I_1_48K_10M - 8, C.RC_I_2_48K_10M - 8 ]
    }

    x = C.X_HAT_F_48K_10M
    x = lc3.tns_synthesize(dt, sr, side, x)
    ok = ok and np.amax(np.abs(x - C.X_HAT_TNS_48K_10M)) < 1e-3

    return ok

def check():

    rng = np.random.default_rng(1234)
    ok = True

    for dt in range(T.NUM_DT):
        for sr in range(T.SRATE_8K, T.SRATE_48K + 1):
            ok = ok and check_analysis(rng, dt, sr)
            ok = ok and check_synthesis(rng, dt, sr)

    for dt in ( T.DT_2M5, T.DT_5M, T.DT_10M ):
        for sr in ( T.SRATE_48K_HR, T.SRATE_96K_HR ):
            ok = ok and check_analysis(rng, dt, sr)

    for dt in ( T.DT_7M5, T.DT_10M ):
        check_analysis_appendix_c(dt)
        check_synthesis_appendix_c(dt)

    return ok

### ------------------------------------------------------------------------ ###



#!/usr/bin/env python3
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mdct, energy, bwdet, attdet
import ltpf, sns, tns, spec, encoder, decoder

ok = True

for m in [ ( mdct    , "MDCT"                   ),
           ( energy  , "Energy Band"            ),
           ( bwdet   , "Bandwidth Detector"     ),
           ( attdet  , "Attack Detector"        ),
           ( ltpf    , "Long Term Postfilter"   ),
           ( sns     , "Spectral Noise Shaping" ),
           ( tns     , "Temporal Noise Shaping" ),
           ( spec    , "Spectral Quantization"  ),
           ( encoder , "Encoder"                ),
           ( decoder , "Decoder"                ) ]:

    print('[{:^6}] {:}'.format('...', m[1]), end='\r', flush=True)
    ret = m[0].check()
    print('[{:^6}] {:}'.format('OK' if ret else 'FAILED', m[1]))

    ok = ok and ret

exit(0 if ok else 1)



#!/usr/bin/env python3
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import lc3
import struct
import sys
import wave

parser = argparse.ArgumentParser(description='LC3 Encoder')

parser.add_argument(
    'wav_file', nargs='?',
    help='Input wave file, default is stdin',
    type=argparse.FileType('rb'), default=sys.stdin.buffer)

parser.add_argument(
    'lc3_file', nargs='?',
    help='Output bitstream file, default is stdout',
    type=argparse.FileType('wb'), default=sys.stdout.buffer)

parser.add_argument(
    '--bitrate', help='Bitrate in bps', type=int, required=True)

parser.add_argument(
    '--frame_duration', help='Frame duration in ms', type=float, default=10)

parser.add_argument(
    '--libpath', help='LC3 Library path')

args = parser.parse_args()

# --- WAV File input ---

f_wav = args.wav_file
wavfile = wave.open(f_wav, 'rb')

samplerate = wavfile.getframerate()
nchannels = wavfile.getnchannels()
bit_depth = wavfile.getsampwidth() * 8
stream_length = wavfile.getnframes()

# --- Setup encoder ---

enc = lc3.Encoder(
    int(args.frame_duration)*1000, samplerate, nchannels, libpath=args.libpath)
frame_size = enc.get_frame_bytes(args.bitrate)
frame_length = enc.get_frame_samples()
bitrate = enc.resolve_bitrate(frame_size)

# --- Setup output ---

f_lc3 = args.lc3_file
f_lc3.write(struct.pack(
    '=HHHHHHHI', 0xcc1c, 18,
    samplerate // 100, bitrate // 100, nchannels,
    int(args.frame_duration * 100), 0, stream_length))

# --- Encoding loop ---

for i in range(0, stream_length, frame_length):

    f_lc3.write(struct.pack('=H', frame_size))

    pcm = wavfile.readframes(frame_length)
    f_lc3.write(enc.encode(pcm, frame_size, bit_depth=bit_depth))

# --- Cleanup ---

wavfile.close()

for f in (f_wav, f_lc3):
    if f is not sys.stdout.buffer:
        f.close()



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

import lc3
import tables as T, appendix_c as C

### ------------------------------------------------------------------------ ###

class EnergyBand:

    def __init__(self, dt, sr):

        self.dt = dt
        self.I = T.I[dt][sr]

    def compute(self, x):

        e = [ np.mean(np.square(x[self.I[i]:self.I[i+1]]))
               for i in range(len(self.I)-1) ]

        e_lo = np.sum(e[:len(e) - [2, 3, 4, 2][self.dt]])
        e_hi = np.sum(e[len(e) - [2, 3, 4, 2][self.dt]:])

        return e, (e_hi > 30*e_lo)

### ------------------------------------------------------------------------ ###

def check_unit(rng, dt, sr):

    ns = T.NS[dt][sr]
    ok = True

    nrg = EnergyBand(dt, sr)

    x = (2 * rng.random(T.NS[dt][sr])) - 1

    (e  , nn  ) = nrg.compute(x)
    (e_c, nn_c) = lc3.energy_compute(dt, sr, x)
    ok = ok and np.amax(np.abs(e_c - e)) < 1e-5 and nn_c == nn

    x[15*ns//16:] *= 1e2

    (e  , nn  ) = nrg.compute(x)
    (e_c, nn_c) = lc3.energy_compute(dt, sr, x)
    ok = ok and np.amax(np.abs(e_c - e)) < 1e-3 and nn_c == nn

    return ok

def check_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    e  = lc3.energy_compute(dt, sr, C.X[i0][0])[0]
    ok = ok and np.amax(np.abs(1 - e/C.E_B[i0][0])) < 1e-6

    e  = lc3.energy_compute(dt, sr, C.X[i0][1])[0]
    ok = ok and np.amax(np.abs(1 - e/C.E_B[i0][1])) < 1e-6

    return ok

def check():

    rng = np.random.default_rng(1234)

    ok = True

    for dt in range(T.NUM_DT):
        for sr in range(T.SRATE_8K, T.SRATE_48K + 1):
            ok = ok and check_unit(rng, dt, sr)

    for dt in ( T.DT_2M5, T.DT_5M, T.DT_10M ):
        for sr in ( T.SRATE_48K_HR, T.SRATE_96K_HR ):
            ok = ok and check_unit(rng, dt, sr)

    for dt in ( T.DT_7M5, T.DT_10M ):
        ok = ok and check_appendix_c(dt)

    return ok

### ------------------------------------------------------------------------ ###



#!/usr/bin/env python3
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import lc3
import struct
import sys
import wave

parser = argparse.ArgumentParser(description='LC3 Decoder')

parser.add_argument(
    'lc3_file', nargs='?',
    help='Input bitstream file, default is stdin',
    type=argparse.FileType('rb'), default=sys.stdin.buffer)

parser.add_argument(
    'wav_file', nargs='?',
    help='Output wave file, default is stdout',
    type=argparse.FileType('wb'), default=sys.stdout.buffer)

parser.add_argument(
    '--bit_depth',
    help='Output bit depth, default is 16 bits',
    type=int, choices=[16, 24], default=16)

parser.add_argument(
    '--libpath', help='LC3 Library path')

args = parser.parse_args()

# --- LC3 File input ---

f_lc3 = args.lc3_file

header = struct.unpack('=HHHHHHHI', f_lc3.read(18))
if header[0] != 0xcc1c:
    raise ValueError('Invalid bitstream file')

samplerate = header[2] * 100
nchannels = header[4]
frame_duration = header[5] * 10
stream_length = header[7]

# --- Setup output ---

bit_depth = args.bit_depth
pcm_size = nchannels * (bit_depth // 8)

f_wav = args.wav_file
wavfile = wave.open(f_wav)
wavfile.setnchannels(nchannels)
wavfile.setsampwidth(bit_depth // 8)
wavfile.setframerate(samplerate)
wavfile.setnframes(stream_length)

# --- Setup decoder ---

dec = lc3.Decoder(
    frame_duration, samplerate, nchannels, libpath=args.libpath)
frame_length = dec.get_frame_samples()
encoded_length = stream_length + dec.get_delay_samples()

# --- Decoding loop ---

for i in range(0, encoded_length, frame_length):

    lc3_frame_size = struct.unpack('=H', f_lc3.read(2))[0]
    pcm = dec.decode(f_lc3.read(lc3_frame_size), bit_depth=bit_depth)

    pcm = pcm[max(encoded_length - stream_length - i, 0) * pcm_size:
              min(encoded_length - i, frame_length) * pcm_size]

    wavfile.writeframesraw(pcm)

# --- Cleanup ---

wavfile.close()

for f in (f_lc3, f_wav):
    if f is not sys.stdout.buffer:
        f.close()



#!/usr/bin/env python3
#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup, Extension
import os, sys, glob
import numpy

if len(sys.argv) <= 1:
  sys.argv = sys.argv + [
    'build', '--build-platlib', os.getcwd() + os.sep + 'build' ]

INC_DIR = '..' + os.sep + 'include'
SRC_DIR = '..' + os.sep + 'src'

sources = glob.glob('*_py.c') + \
          [ SRC_DIR + os.sep + 'tables.c',
            SRC_DIR + os.sep + 'bits.c',
            SRC_DIR + os.sep + 'plc.c' ]

depends = [ 'ctypes.h' ] + \
          glob.glob(INC_DIR + os.sep + '*.h') + \
          glob.glob(SRC_DIR + os.sep + '*.[c,h]')

includes = [ SRC_DIR, INC_DIR, numpy.get_include() ]

extension = Extension('lc3',
  extra_compile_args = [ '-std=c11', '-ffast-math' ],
  define_macros = [ ('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION') ],
  sources = sources,
  depends = depends,
  include_dirs = includes)

setup(name = 'LC3',
      version = '1.0',
      description = 'LC3 Test Python Module',
      ext_modules = [ extension ])



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

import lc3
import tables as T, appendix_c as C


BW_START = [
    [ [], [ 24 ], [ 24, 35 ], [ 24, 33, 39 ], [ 22, 31, 37, 41 ] ],
    [ [], [ 39 ], [ 35, 47 ], [ 34, 44, 50 ], [ 32, 42, 48, 52 ] ],
    [ [], [ 51 ], [ 45, 58 ], [ 42, 53, 60 ], [ 40, 51, 57, 61 ] ],
    [ [], [ 53 ], [ 47, 59 ], [ 44, 54, 60 ], [ 41, 51, 57, 61 ] ]
]

BW_STOP = [
    [ [], [ 34 ], [ 32, 39 ], [ 31, 38, 42 ], [ 29, 35, 40, 43 ] ],
    [ [], [ 49 ], [ 44, 51 ], [ 42, 49, 53 ], [ 40, 46, 51, 54 ] ],
    [ [], [ 63 ], [ 55, 63 ], [ 51, 58, 63 ], [ 48, 55, 60, 63 ] ],
    [ [], [ 63 ], [ 56, 63 ], [ 52, 59, 63 ], [ 49, 55, 60, 63 ] ]
]

TQ = [ 20, 10, 10, 10 ]

TC = [ 15, 23, 20, 20 ]
L  = [ [ 4, 4, 3, 1 ], [ 4, 4, 3, 1 ],
       [ 4, 4, 3, 2 ], [ 4, 4, 3, 1 ] ]

### ------------------------------------------------------------------------ ###

class BandwidthDetector:

    def __init__(self, dt, sr):

        self.dt = dt
        self.sr = sr

    def run(self, e):

        dt = self.dt
        sr = self.sr

        ### Stage 1, determine bw0 candidate

        bw0 = 0

        for bw in range(sr):
            i0 = BW_START[dt][sr][bw]
            i1 = BW_STOP[dt][sr][bw]
            if np.mean(e[i0:i1+1]) >= TQ[bw]:
                bw0 = bw + 1

        ### Stage 2, Cut-off random coefficients at each steps

        bw = bw0

        if bw0 < sr:
            l  = L[dt][bw0]
            i0 = BW_START[dt][sr][bw0] - l
            i1 = BW_START[dt][sr][bw0]

            c = 10 * np.log10(1e-31 + e[i0-l+1:i1-l+2] / e[i0+1:i1+2])
            if np.amax(c) <= TC[bw0]:
                bw = sr

        self.bw = bw
        return self.bw

    def get_nbits(self):

        return 0 if self.sr == 0 else \
               1 + np.log2(self.sr).astype(int)

    def store(self, b):

        b.write_uint(self.bw, self.get_nbits())

    def get(self, b):

        return b.read_uint(self.get_nbits())

### ------------------------------------------------------------------------ ###

def check_unit(rng, dt, sr):

    ok = True

    bwdet = BandwidthDetector(dt, sr)

    for bw0 in range(sr+1):
        for drop in range(10):

            ### Generate random 'high' energy and
            ### scale relevant bands to select 'bw0'

            e = 20 + 100 * rng.random(64)

            for i in range(sr):
                if i+1 != bw0:
                    i0 = BW_START[dt][sr][i]
                    i1 = BW_STOP[dt][sr][i]
                    e[i0:i1+1] /= (np.mean(e[i0:i1+1]) / TQ[i] + 1e-3)

            ### Stage 2 Condition,
            ### cut-off random coefficients at each steps

            if bw0 < sr:
                l  = L[dt][bw0]
                i0 = BW_START[dt][sr][bw0] - l
                i1 = BW_START[dt][sr][bw0]

                e[i0-l+1:i1+2] /= np.power(10, np.arange(2*l+1) / (1 + drop))

            ### Check with implementation

            bw_c = lc3.bwdet_run(dt, sr, e)

            ok = ok and bw_c == bwdet.run(e)

    return ok

def check_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    E_B  = C.E_B[i0]
    P_BW = C.P_BW[i0]

    bw = lc3.bwdet_run(dt, sr, E_B[0])
    ok = ok and bw == P_BW[0]

    bw = lc3.bwdet_run(dt, sr, E_B[1])
    ok = ok and bw == P_BW[1]

    return ok

def check():

    rng = np.random.default_rng(1234)

    ok = True
    for dt in range(T.NUM_DT):
        for sr in range(T.SRATE_8K, T.SRATE_48K + 1):
            ok = ok and check_unit(rng, dt, sr)

    for dt in ( T.DT_7M5, T.DT_10M ):
        ok = ok and check_appendix_c(dt)

    return ok

### ------------------------------------------------------------------------ ###



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import scipy.fftpack as fftpack

import lc3
import tables as T, appendix_c as C

### ------------------------------------------------------------------------ ###

class Sns:

    def __init__(self, dt, sr):

        self.dt = dt
        self.sr = sr
        self.I = T.I[dt][sr]

        (self.ind_lf, self.ind_hf, self.shape, self.gain) = \
            (None, None, None, None)

        (self.idx_a, self.ls_a, self.idx_b, self.ls_b) = \
            (None, None, None, None)

    def get_data(self):

        data = { 'lfcb' : self.ind_lf, 'hfcb' : self.ind_hf,
                 'shape' : self.shape, 'gain' : self.gain,
                 'idx_a' : self.idx_a, 'ls_a' : self.ls_a }

        if self.idx_b is not None:
            data.update({ 'idx_b' : self.idx_b, 'ls_b' : self.ls_b })

        return data

    def get_nbits(self):

        return 38

    def spectral_shaping(self, scf, inv, x):

        ## Scale factors interpolation

        scf_i = np.empty(4*len(scf))
        scf_i[0     ] = scf[0]
        scf_i[1     ] = scf[0]
        scf_i[2:62:4] = scf[:15] + 1/8 * (scf[1:] - scf[:15])
        scf_i[3:63:4] = scf[:15] + 3/8 * (scf[1:] - scf[:15])
        scf_i[4:64:4] = scf[:15] + 5/8 * (scf[1:] - scf[:15])
        scf_i[5:64:4] = scf[:15] + 7/8 * (scf[1:] - scf[:15])
        scf_i[62    ] = scf[15 ] + 1/8 * (scf[15] - scf[14 ])
        scf_i[63    ] = scf[15 ] + 3/8 * (scf[15] - scf[14 ])

        nb = len(self.I) - 1

        if nb < 32:
            n4 = round(abs(1-32/nb)*nb)
            n2 = nb - n4

            for i in range(n4):
                scf_i[i] = np.mean(scf_i[4*i:4*i+4])

            for i in range(n4, n4+n2):
                scf_i[i] = np.mean(scf_i[2*n4+2*i:2*n4+2*i+2])

            scf_i = scf_i[:n4+n2]

        elif nb < 64:
            n2 = 64 - nb

            for i in range(n2):
                scf_i[i] = np.mean(scf_i[2*i:2*i+2])
            scf_i = np.append(scf_i[:n2], scf_i[2*n2:])

        g_sns = np.power(2, [ -scf_i, scf_i ][inv])

        ## Spectral shaping

        y = np.empty(len(x))
        I = self.I

        for b in range(nb):
            y[I[b]:I[b+1]] = x[I[b]:I[b+1]] * g_sns[b]

        return y


class SnsAnalysis(Sns):

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

    def compute_scale_factors(self, e, att, nbytes):

        dt = self.dt
        sr = self.sr
        hr = self.sr >= T.SRATE_48K_HR

        ## Padding

        if len(e) < 32:
            n4 = round(abs(1-32/len(e))*len(e))
            n2 = len(e) - n4

            e = np.append(np.zeros(3*n4+n2), e)
            for i in range(n4):
                e[4*i+0] = e[4*i+1] = \
                e[4*i+2] = e[4*i+3] = e[3*n4+n2+i]

            for i in range(2*n4, 2*n4+n2):
                e[2*i+0] = e[2*i+1] = e[2*n4+n2+i]

        elif len(e) < 64:
            n2 = 64 - len(e)

            e = np.append(np.empty(n2), e)
            for i in range(n2):
                e[2*i+0] = e[2*i+1] = e[n2+i]

        ## Smoothing

        e_s = np.zeros(len(e))
        e_s[0   ] = 0.75 * e[0   ] + 0.25 * e[1   ]
        e_s[1:63] = 0.25 * e[0:62] + 0.5  * e[1:63] + 0.25 * e[2:64]
        e_s[  63] = 0.25 * e[  62] + 0.75 * e[  63]

        ## Pre-emphasis

        g_tilt = [ 14, 18, 22, 26, 30, 30, 34 ][self.sr]
        e_p = e_s * (10 ** ((np.arange(64) * g_tilt) / 630))

        ## Noise floor

        noise_floor = max(np.average(e_p) * (10 ** (-40/10)), 2 ** -32)
        e_p = np.fmax(e_p, noise_floor * np.ones(len(e)))

        ## Logarithm

        e_l = np.log2(10 ** -31 + e_p) / 2

        ## Band energy grouping

        w = [ 1/12, 2/12, 3/12, 3/12, 2/12, 1/12 ]

        e_4 = np.zeros(len(e_l) // 4)
        e_4[0   ] = w[0] * e_l[0] + np.sum(w[1:] * e_l[:5])
        e_4[1:15] = [ np.sum(w * e_l[4*i-1:4*i+5]) for i in range(1, 15) ]
        e_4[  15] = np.sum(w[:5] * e_l[59:64]) + w[5] * e_l[63]

        ## Mean removal and scaling, attack handling

        cf = [ 0.85, 0.6 ][hr]
        if hr and nbytes * 8 > [ 1150, 2300, 0, 4400 ][self.dt]:
            cf *= [ 0.25, 0.35 ][ self.dt == T.DT_10M ]

        scf = cf * (e_4 - np.average(e_4))

        scf_a = np.zeros(len(scf))
        scf_a[0   ] = np.mean(scf[:3])
        scf_a[1   ] = np.mean(scf[:4])
        scf_a[2:14] = [ np.mean(scf[i:i+5]) for i in range(12) ]
        scf_a[  14] = np.mean(scf[12:])
        scf_a[  15] = np.mean(scf[13:])

        scf_a = (0.5 if self.dt != T.DT_7M5 else 0.3) * \
                (scf_a - np.average(scf_a))

        return scf_a if att else scf

    def enum_mpvq(self, v):

        sign = None
        index = 0
        x = 0

        for (n, vn) in enumerate(v[::-1]):

            if sign is not None and vn != 0:
                index = 2*index + sign
            if vn != 0:
                sign = 1 if vn < 0 else 0

            index += T.SNS_MPVQ_OFFSETS[n][x]
            x += abs(vn)

        return (index, bool(sign))

    def quantize(self, scf):

        ## Stage 1

        dmse_lf = [ np.sum((scf[:8] - T.SNS_LFCB[i]) ** 2) for i in range(32) ]
        dmse_hf = [ np.sum((scf[8:] - T.SNS_HFCB[i]) ** 2) for i in range(32) ]

        self.ind_lf = np.argmin(dmse_lf)
        self.ind_hf = np.argmin(dmse_hf)

        st1 = np.append(T.SNS_LFCB[self.ind_lf], T.SNS_HFCB[self.ind_hf])
        r1 = scf - st1

        ## Stage 2

        t2_rot = fftpack.dct(r1, norm = 'ortho')
        x = np.abs(t2_rot)

        ## Stage 2 Shape search, step 1

        K = 6

        proj_fac = (K - 1) / sum(np.abs(t2_rot))
        y3 = np.floor(x * proj_fac).astype(int)

        ## Stage 2 Shape search, step 2

        corr_xy = np.sum(y3 * x)
        energy_y = np.sum(y3 * y3)

        k0 = sum(y3)
        for k in range(k0, K):
            q_pvq = ((corr_xy + x) ** 2) / (energy_y + 2*y3 + 1)
            n_best = np.argmax(q_pvq)

            corr_xy += x[n_best]
            energy_y += 2*y3[n_best] + 1
            y3[n_best] += 1

        ## Stage 2 Shape search, step 3

        K = 8

        y2 = y3.copy()

        for k in range(sum(y2), K):
            q_pvq = ((corr_xy + x) ** 2) / (energy_y + 2*y2 + 1)
            n_best = np.argmax(q_pvq)

            corr_xy += x[n_best]
            energy_y += 2*y2[n_best] + 1
            y2[n_best] += 1


        ## Stage 2 Shape search, step 4

        y1 = np.append(y2[:10], [0] * 6)

        ## Stage 2 Shape search, step 5

        corr_xy -= sum(y2[10:] * x[10:])
        energy_y -= sum(y2[10:] * y2[10:])

        ## Stage 2 Shape search, step 6

        K = 10

        for k in range(sum(y1), K):
            q_pvq = ((corr_xy + x[:10]) ** 2) / (energy_y + 2*y1[:10] + 1)
            n_best = np.argmax(q_pvq)

            corr_xy += x[n_best]
            energy_y += 2*y1[n_best] + 1
            y1[n_best] += 1

        ## Stage 2 Shape search, step 7

        y0 = np.append(y1[:10], [ 0 ] * 6)

        q_pvq = ((corr_xy + x[10:]) ** 2) / (energy_y + 2*y0[10:] + 1)
        n_best = 10 + np.argmax(q_pvq)

        y0[n_best] += 1

        ## Stage 2 Shape search, step 8

        y0 *= np.sign(t2_rot).astype(int)
        y1 *= np.sign(t2_rot).astype(int)
        y2 *= np.sign(t2_rot).astype(int)
        y3 *= np.sign(t2_rot).astype(int)

        ## Stage 2 Shape search, step 9

        xq = [ y / np.sqrt(sum(y ** 2)) for y in (y0, y1, y2, y3) ]

        ## Shape and gain combination determination

        G = [ T.SNS_VQ_REG_ADJ_GAINS, T.SNS_VQ_REG_LF_ADJ_GAINS,
              T.SNS_VQ_NEAR_ADJ_GAINS, T.SNS_VQ_FAR_ADJ_GAINS ]

        dMSE = [ [ sum((t2_rot - G[j][i] * xq[j]) ** 2)
                   for i in range(len(G[j])) ] for j in range(4) ]

        self.shape = np.argmin([ np.min(dMSE[j]) for j in range(4) ])
        self.gain = np.argmin(dMSE[self.shape])

        gain = G[self.shape][self.gain]

        ## Enumeration of the selected PVQ pulse configurations

        if self.shape == 0:
            (self.idx_a, self.ls_a) = self.enum_mpvq(y0[:10])
            (self.idx_b, self.ls_b) = self.enum_mpvq(y0[10:])
        elif self.shape == 1:
            (self.idx_a, self.ls_a) = self.enum_mpvq(y1[:10])
            (self.idx_b, self.ls_b) = (None, None)
        elif self.shape == 2:
            (self.idx_a, self.ls_a) = self.enum_mpvq(y2)
            (self.idx_b, self.ls_b) = (None, None)
        elif self.shape == 3:
            (self.idx_a, self.ls_a) = self.enum_mpvq(y3)
            (self.idx_b, self.ls_b) = (None, None)

        ## Synthesis of the Quantized scale factor

        scf_q = st1 + gain * fftpack.idct(xq[self.shape], norm = 'ortho')

        return scf_q

    def run(self, eb, att, nbytes, x):

        scf = self.compute_scale_factors(eb, att, nbytes)
        scf_q = self.quantize(scf)
        y = self.spectral_shaping(scf_q, False, x)

        return y

    def store(self, b):

        shape = self.shape
        gain_msb_bits = np.array([ 1, 1, 2, 2 ])[shape]
        gain_lsb_bits = np.array([ 0, 1, 0, 1 ])[shape]

        b.write_uint(self.ind_lf, 5)
        b.write_uint(self.ind_hf, 5)

        b.write_bit(shape >> 1)

        b.write_uint(self.gain >> gain_lsb_bits, gain_msb_bits)

        b.write_bit(self.ls_a)

        if self.shape == 0:
            sz_shape_a = 2390004
            index_joint = self.idx_a + \
                (2 * self.idx_b + self.ls_b + 2) * sz_shape_a

        elif self.shape == 1:
            sz_shape_a = 2390004
            index_joint = self.idx_a + (self.gain & 1) * sz_shape_a

        elif self.shape == 2:
            index_joint = self.idx_a

        elif self.shape == 3:
            sz_shape_a = 15158272
            index_joint = sz_shape_a + (self.gain & 1) + 2 * self.idx_a

        b.write_uint(index_joint, 14 - gain_msb_bits)
        b.write_uint(index_joint >> (14 - gain_msb_bits), 12)


class SnsSynthesis(Sns):

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

    def deenum_mpvq(self, index, ls, npulses, n):

        y = np.zeros(n, dtype=np.intc)
        pos = 0

        for i in range(len(y)-1, -1, -1):

            if index > 0:
                yi = 0
                while index < T.SNS_MPVQ_OFFSETS[i][npulses - yi]: yi += 1
                index -= T.SNS_MPVQ_OFFSETS[i][npulses - yi]
            else:
                yi = npulses

            y[pos] = [ yi, -yi ][int(ls)]
            pos += 1

            npulses -= yi
            if npulses <= 0:
                break

            if yi > 0:
                ls = index & 1
                index >>= 1

        return y

    def unquantize(self):

        ## SNS VQ Decoding

        y = np.empty(16, dtype=np.intc)

        if self.shape == 0:
            y[:10] = self.deenum_mpvq(self.idx_a, self.ls_a, 10, 10)
            y[10:] = self.deenum_mpvq(self.idx_b, self.ls_b,  1,  6)
        elif self.shape == 1:
            y[:10] = self.deenum_mpvq(self.idx_a, self.ls_a, 10, 10)
            y[10:] = np.zeros(6, dtype=np.intc)
        elif self.shape == 2:
            y = self.deenum_mpvq(self.idx_a, self.ls_a, 8, 16)
        elif self.shape == 3:
            y = self.deenum_mpvq(self.idx_a, self.ls_a, 6, 16)

        ## Unit energy normalization

        y = y / np.sqrt(sum(y ** 2))

        ## Reconstruction of the quantized scale factors

        G = [ T.SNS_VQ_REG_ADJ_GAINS, T.SNS_VQ_REG_LF_ADJ_GAINS,
              T.SNS_VQ_NEAR_ADJ_GAINS, T.SNS_VQ_FAR_ADJ_GAINS ]

        gain = G[self.shape][self.gain]

        scf = np.append(T.SNS_LFCB[self.ind_lf], T.SNS_HFCB[self.ind_hf]) \
                + gain * fftpack.idct(y, norm = 'ortho')

        return scf

    def load(self, b):

        self.ind_lf = b.read_uint(5)
        self.ind_hf = b.read_uint(5)

        shape_msb = b.read_bit()

        gain_msb_bits = 1 + shape_msb
        self.gain = b.read_uint(gain_msb_bits)

        self.ls_a = b.read_bit()

        index_joint  = b.read_uint(14 - gain_msb_bits)
        index_joint |= b.read_uint(12) << (14 - gain_msb_bits)

        if shape_msb == 0:
            sz_shape_a = 2390004

            if index_joint >= sz_shape_a * 14:
                raise ValueError('Invalide SNS joint index')

            self.idx_a = index_joint % sz_shape_a
            index_joint = index_joint // sz_shape_a
            if index_joint >= 2:
                self.shape = 0
                self.idx_b = (index_joint - 2) // 2
                self.ls_b =  (index_joint - 2)  % 2
            else:
                self.shape = 1
                self.gain = (self.gain << 1) + (index_joint & 1)

        else:
            sz_shape_a = 15158272
            if index_joint >= sz_shape_a + 1549824:
                raise ValueError('Invalide SNS joint index')

            if index_joint < sz_shape_a:
                self.shape = 2
                self.idx_a = index_joint
            else:
                self.shape = 3
                index_joint -= sz_shape_a
                self.gain = (self.gain << 1) + (index_joint % 2)
                self.idx_a = index_joint // 2

    def run(self, x):

        scf = self.unquantize()
        y = self.spectral_shaping(scf, True, x)

        return y

### ------------------------------------------------------------------------ ###

def check_analysis(rng, dt, sr):

    ok = True

    analysis = SnsAnalysis(dt, sr)

    for i in range(10):
        ne = T.I[dt][sr][-1]
        x  = rng.random(ne) * 1e4
        e  = rng.random(len(T.I[dt][sr]) - 1) * 1e10

        if sr >= T.SRATE_48K_HR:
            for nbits in (1144, 1152, 2296, 2304, 4400, 4408):
                y = analysis.run(e, False, nbits // 8, x)
                data = analysis.get_data()

                (y_c, data_c) = lc3.sns_analyze(
                    dt, sr, nbits // 8, e, False, x)

                for k in data.keys():
                    ok = ok and data_c[k] == data[k]

                ok = ok and lc3.sns_get_nbits() == analysis.get_nbits()
                ok = ok and np.amax(np.abs(y - y_c)) < 1e-1

        else:
            for att in (0, 1):
                y = analysis.run(e, att, 0, x)
                data = analysis.get_data()

                (y_c, data_c) = lc3.sns_analyze(dt, sr, 0, e, att, x)

                for k in data.keys():
                    ok = ok and data_c[k] == data[k]

                ok = ok and lc3.sns_get_nbits() == analysis.get_nbits()
                ok = ok and np.amax(np.abs(y - y_c)) < 1e-1

    return ok

def check_synthesis(rng, dt, sr):

    ok = True

    synthesis = SnsSynthesis(dt, sr)

    for i in range(100):

        synthesis.ind_lf = rng.integers(0, 32)
        synthesis.ind_hf = rng.integers(0, 32)

        shape = rng.integers(0, 4)
        sz_shape_a = [ 2390004, 2390004, 15158272, 774912 ][shape]
        sz_shape_b = [ 6, 1, 0, 0 ][shape]
        synthesis.shape = shape
        synthesis.gain = rng.integers(0, [ 2, 4, 4, 8 ][shape])
        synthesis.idx_a = rng.integers(0, sz_shape_a, endpoint=True)
        synthesis.ls_a = bool(rng.integers(0, 1, endpoint=True))
        synthesis.idx_b = rng.integers(0, sz_shape_b, endpoint=True)
        synthesis.ls_b = bool(rng.integers(0, 1, endpoint=True))

        ne = T.I[dt][sr][-1]
        x  = rng.random(ne) * 1e4

        y = synthesis.run(x)
        y_c = lc3.sns_synthesize(dt, sr, synthesis.get_data(), x)
        ok = ok and np.amax(np.abs(1 - y/y_c)) < 1e-5

    return ok

def check_analysis_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    for i in range(len(C.E_B[i0])):

        scf = lc3.sns_compute_scale_factors(dt, sr, 0, C.E_B[i0][i], False)
        ok = ok and np.amax(np.abs(scf - C.SCF[i0][i])) < 1e-4

        (lf, hf) = lc3.sns_resolve_codebooks(scf)
        ok = ok and lf == C.IND_LF[i0][i] and hf == C.IND_HF[i0][i]

        (y, yn, shape, gain) = lc3.sns_quantize(scf, lf, hf)
        ok = ok and np.any(y[0][:16] - C.SNS_Y0[i0][i] == 0)
        ok = ok and np.any(y[1][:10] - C.SNS_Y1[i0][i] == 0)
        ok = ok and np.any(y[2][:16] - C.SNS_Y2[i0][i] == 0)
        ok = ok and np.any(y[3][:16] - C.SNS_Y3[i0][i] == 0)
        ok = ok and shape == 2*C.SUBMODE_MSB[i0][i] + C.SUBMODE_LSB[i0][i]
        ok = ok and gain == C.G_IND[i0][i]

        scf_q = lc3.sns_unquantize(lf, hf, yn[shape], shape, gain)
        ok = ok and np.amax(np.abs(scf_q - C.SCF_Q[i0][i])) < 1e-5

        x = lc3.sns_spectral_shaping(dt, sr, C.SCF_Q[i0][i], False, C.X[i0][i])
        ok = ok and np.amax(np.abs(1 - x/C.X_S[i0][i])) < 1e-5

        (x, data) = lc3.sns_analyze(dt, sr, 0, C.E_B[i0][i], False, C.X[i0][i])
        ok = ok and data['lfcb'] == C.IND_LF[i0][i]
        ok = ok and data['hfcb'] == C.IND_HF[i0][i]
        ok = ok and data['shape'] == 2*C.SUBMODE_MSB[i0][i] + \
                                       C.SUBMODE_LSB[i0][i]
        ok = ok and data['gain'] == C.G_IND[i0][i]
        ok = ok and data['idx_a'] == C.IDX_A[i0][i]
        ok = ok and data['ls_a'] == C.LS_IND_A[i0][i]
        ok = ok and (C.IDX_B[i0][i] is None or
            data['idx_b'] == C.IDX_B[i0][i])
        ok = ok and (C.LS_IND_B[i0][i] is None or
            data['ls_b'] == C.LS_IND_B[i0][i])
        ok = ok and np.amax(np.abs(1 - x/C.X_S[i0][i])) < 1e-5

    return ok

def check_synthesis_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    for i in range(len(C.X_HAT_TNS[i0])):

        data = {
            'lfcb'  : C.IND_LF[i0][i], 'hfcb' : C.IND_HF[i0][i],
            'shape' : 2*C.SUBMODE_MSB[i0][i] + C.SUBMODE_LSB[i0][i],
            'gain'  : C.G_IND[i0][i],
            'idx_a' : C.IDX_A[i0][i],
            'ls_a'  : C.LS_IND_A[i0][i],
            'idx_b' : C.IDX_B[i0][i] if C.IDX_B[i0][i] is not None else 0,
            'ls_b'  : C.LS_IND_B[i0][i] if C.LS_IND_B[i0][i] is not None else 0,
        }

        x = lc3.sns_synthesize(dt, sr, data, C.X_HAT_TNS[i0][i])
        ok = ok and np.amax(np.abs(x - C.X_HAT_SNS[i0][i])) < 1e0

    return ok

def check():

    rng = np.random.default_rng(1234)
    ok = True

    for dt in range(T.NUM_DT):
        for sr in range(T.SRATE_8K, T.SRATE_48K + 1):
            ok = ok and check_analysis(rng, dt, sr)
            ok = ok and check_synthesis(rng, dt, sr)

    for dt in ( T.DT_2M5, T.DT_5M, T.DT_10M ):
        for sr in ( T.SRATE_48K_HR, T.SRATE_96K_HR ):
            ok = ok and check_analysis(rng, dt, sr)
            ok = ok and check_synthesis(rng, dt, sr)

    for dt in ( T.DT_7M5, T.DT_10M ):
        check_analysis_appendix_c(dt)
        check_synthesis_appendix_c(dt)

    return ok

### ------------------------------------------------------------------------ ###



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import scipy.fft

import lc3
import tables as T, appendix_c as C

### ------------------------------------------------------------------------ ###

class Mdct:

    W = [ [ T.W_2M5_8K , T.W_2M5_16K, T.W_2M5_24K,
            T.W_2M5_32K, T.W_2M5_48K, T.W_2M5_48K_HR, T.W_2M5_96K_HR ],

          [ T.W_5M_8K  , T.W_5M_16K , T.W_5M_24K ,
            T.W_5M_32K , T.W_5M_48K , T.W_5M_48K_HR , T.W_5M_96K_HR  ],

          [ T.W_7M5_8K , T.W_7M5_16K, T.W_7M5_24K,
            T.W_7M5_32K, T.W_7M5_48K, None, None ],

          [ T.W_10M_8K , T.W_10M_16K, T.W_10M_24K,
            T.W_10M_32K, T.W_10M_48K, T.W_10M_48K_HR, T.W_10M_96K_HR ] ]

    def __init__(self, dt, sr):

        self.ns = T.NS[dt][sr]
        self.nd = T.ND[dt][sr]

        self.t = np.zeros(2*self.ns)
        self.w = Mdct.W[dt][sr]


class MdctForward(Mdct):

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

    def run(self, x):

        ns = self.ns
        nd = self.nd

        self.t[nd:nd+ns] = x
        t = self.t * self.w
        self.t[0:nd] = x[ns-nd:]

        n  = len(t)
        n2 = n // 2

        z = t * np.exp(-2j * np.pi * np.arange(n) / (2*n))
        z = scipy.fft.fft(z)[:n2]
        z = z * np.exp(-2j * np.pi *
                (n2/2 + 0.5) * (np.arange(n2) + 0.5) / (2 * n2))
        return np.real(z) * np.sqrt(2/n2)


class MdctInverse(Mdct):

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

    def run(self, x):

        ns = self.ns
        nd = self.nd

        n = len(x)

        x = np.append(x, -x[::-1])
        z = x * np.exp(2j * np.pi * (n/2 + 0.5) * np.arange(2*n) / (2*n))
        z = scipy.fft.ifft(z) * n
        z = z * np.exp(2j * np.pi * (np.arange(2*n) + (n/2 + 0.5)) / (4*n))
        t = np.real(z) * np.sqrt(2/n)

        t = t * self.w[::-1]

        y = np.empty(ns)
        y[:nd] = t[ns-nd:ns] + self.t[2*ns-nd:]
        y[nd:] = t[ns:2*ns-nd]
        self.t = t

        return y

### ------------------------------------------------------------------------ ###

def check_forward_unit(rng, dt, sr):

    ns = T.NS[dt][sr]
    nd = T.ND[dt][sr]
    ok = True

    x = (2 * rng.random(ns)) - 1

    y   = [ None ] * 2
    y_c = [ None ] * 2

    mdct = MdctForward(dt, sr)
    y[0] = mdct.run(x)
    y[1] = mdct.run(x)

    (y_c[0], d_c) = lc3.mdct_forward(dt, sr, x, np.zeros(nd))
    y_c[1] = lc3.mdct_forward(dt, sr, x, d_c)[0]

    ok = ok and np.amax(np.abs(y[0] - y_c[0])) < 1e-5
    ok = ok and np.amax(np.abs(y[1] - y_c[1])) < 1e-5

    return ok


def check_forward_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    ns = T.NS[dt][sr]
    nd = T.ND[dt][sr]

    (y, d) = lc3.mdct_forward(dt, sr, C.X_PCM[i0][0], np.zeros(nd))
    ok = ok and np.amax(np.abs(y - C.X[i0][0])) < 1e-1

    (y, d) = lc3.mdct_forward(dt, sr, C.X_PCM[i0][1], d)
    ok = ok and np.amax(np.abs(y - C.X[i0][1])) < 1e-1

    return ok


def check_inverse_unit(rng, dt, sr):

    ns = T.NS[dt][sr]
    nd = T.ND[dt][sr]
    ok = True

    x  = (2 * rng.random(ns)) - 1

    y   = [ None ] * 2
    y_c = [ None ] * 2

    mdct = MdctInverse(dt, sr)
    y[0] = mdct.run(x)
    y[1] = mdct.run(x)

    (y_c[0], d_c) = lc3.mdct_inverse(dt, sr, x, np.zeros(nd))
    y_c[1] = lc3.mdct_inverse(dt, sr, x, d_c)[0]

    ok = ok and np.amax(np.abs(y[0] - y_c[0])) < 1e-5
    ok = ok and np.amax(np.abs(y[1] - y_c[1])) < 1e-5

    return ok


def check_inverse_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    ns = T.NS[dt][sr]
    nd = T.ND[dt][sr]

    (y, d0) = lc3.mdct_inverse(dt, sr, C.X_HAT_SNS[i0][0], np.zeros(nd))
    yr = C.T_HAT_MDCT[i0][0][ns-nd:2*ns-nd]
    dr = C.T_HAT_MDCT[i0][0][2*ns-nd:]

    ok = ok and np.amax(np.abs(yr - y )) < 1e-1
    ok = ok and np.amax(np.abs(dr - d0)) < 1e-1

    (y, d1) = lc3.mdct_inverse(dt, sr, C.X_HAT_SNS[i0][1], d0)
    yr[  :nd] = C.T_HAT_MDCT[i0][1][ns-nd:ns] + d0
    yr[nd:ns] = C.T_HAT_MDCT[i0][1][ns:2*ns-nd]
    dr        = C.T_HAT_MDCT[i0][1][2*ns-nd:]

    ok = ok and np.amax(np.abs(yr - y )) < 1e-1
    ok = ok and np.amax(np.abs(dr - d1)) < 1e-1

    return ok


def check():

    rng = np.random.default_rng(1234)

    ok  = True

    for dt in range(T.NUM_DT):
        for sr in range(T.SRATE_8K, T.SRATE_48K + 1):
            ok = ok and check_forward_unit(rng, dt, sr)
            ok = ok and check_inverse_unit(rng, dt, sr)

    for dt in ( T.DT_2M5, T.DT_5M, T.DT_10M ):
        for sr in ( T.SRATE_48K_HR, T.SRATE_96K_HR ):
            ok = ok and check_forward_unit(rng, dt, sr)
            ok = ok and check_inverse_unit(rng, dt, sr)

    for dt in ( T.DT_7M5, T.DT_10M ):
        ok = ok and check_forward_appendix_c(dt)
        ok = ok and check_inverse_appendix_c(dt)

    return ok



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

import lc3
import tables as T, appendix_c as C

import bwdet as m_bwdet
import ltpf as m_ltpf
import sns as m_sns
import tns as m_tns

### ------------------------------------------------------------------------ ###

class SpectrumQuantization:

    def __init__(self, dt, sr):

        self.dt = dt
        self.sr = sr

    def get_gain_offset(self, nbytes):

        sr_ind = self.sr if self.sr < T.SRATE_48K_HR \
            else 4 + (self.sr - T.SRATE_48K_HR)

        g_off = (nbytes * 8) // (10 * (1 + sr_ind))
        g_off = -min(115, g_off) - (105 + 5*(1 + sr_ind))
        if self.sr >= T.SRATE_48K_HR:
            g_off = max(g_off, -181)

        return g_off

    def get_noise_indices(self, bw, xq, lastnz):

        nf_start = [  6, 12, 18, 24 ][self.dt]
        nf_width = [  1,  1,  2,  3 ][self.dt]

        bw_stop = T.I[self.dt][min(bw, T.SRATE_48K)][-1]

        xq = np.append(xq[:lastnz], np.zeros(len(xq) - lastnz))
        xq[:nf_start-nf_width] = 1

        return [ np.all(xq[max(k-nf_width, 0):min(k+nf_width+1, bw_stop)] == 0)
                    for k in range(bw_stop) ]

class SpectrumAnalysis(SpectrumQuantization):

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

        self.reset_off  = 0
        self.nbits_off  = 0
        self.nbits_spec = 0
        self.nbits_est  = 0

        (self.g_idx, self.noise_factor, self.xq, self.lastnz,
                self.nbits_residual_max, self.xg) = \
            (None, None, None, None, None, None)

    def estimate_gain(self, x, nbytes, nbits_spec, nbits_off, g_off):

        nbits = int(nbits_spec + nbits_off + 0.5)

        ### Energy (dB) by 4 MDCT coefficients

        hr = (self.sr >= T.SRATE_48K_HR)
        nf = 0

        if hr:
            dt = self.dt
            sr = self.sr

            dt_ms = T.DT_MS[dt]
            bitrate = (8 * nbytes / (dt_ms * 1e-3)).astype(int)

            C = [ [ -6, 0, None, 2 ], [ -6, 0, None, 5 ] ]

            reg_bits = np.clip(
                bitrate // 12500 + C[sr - T.SRATE_48K_HR][dt], 6, 23)

            M0 = np.sum(np.abs(x)) + 1e-5
            M1 = np.sum(np.arange(len(x)) * np.abs(x)) + 1e-5

            low_bits = (4 / dt_ms) * (2*dt_ms - min(M0/M1, 2*dt_ms))

            nf = np.max(np.abs(x)) * np.exp2(-reg_bits - low_bits)

        e = [ np.sum(x[4*k:4*(k+1)] ** 2) for k in range(len(x) // 4) ]
        e = 10 * np.log10(2**-31 + np.array(e) + nf)

        ### Compute gain index

        g_idx = 255

        for i in range(8):
            factor = 1 << (7 - i)
            g_idx -= factor
            tmp = 0
            iszero = 1

            for ei in e[-1::-1]:

                if ei * 28/20 < g_idx + g_off:
                    if iszero == 0:
                        tmp += 2.7*28/20
                else:
                    if g_idx + g_off < (ei - 43) * 28/20:
                        tmp += 2*ei*28/20 - 2*(g_idx + g_off) - 36*28/20
                    else:
                        tmp += ei*28/20 - (g_idx + g_off) + 7*28/20
                    iszero = 0

            if tmp > nbits * 1.4 * 28/20 and iszero == 0:
                g_idx += factor

        ### Limit gain index

        x_max = np.amax(np.abs(x))
        if x_max > 0:
            x_lim = [ 2**15 - 0.375,  2**23 ][hr]
            g_min = 28 * np.log10(x_max / x_lim)
            g_min = np.ceil(g_min).astype(int) - g_off
            reset_off = g_idx < g_min
        else:
            g_min = 0
            reset_off = True

        if reset_off:
            g_idx = g_min

        return (g_min, g_idx + g_off, reset_off)

    def quantize(self, g_int, x):

        xg = x / 10 ** (g_int / 28)

        hr = (self.sr >= T.SRATE_48K_HR)
        offset = [ 0.375, 0.5 ][hr]
        xq_min = [ -(2**15)  , -(2**23)   ][hr]
        xq_max = [  (2**15)-1,  (2**23)-1 ][hr]

        xq = np.where(xg < 0, np.ceil(xg - offset), np.floor(xg + offset))
        xq = xq.astype(np.int32)
        xq = np.fmin(np.fmax(xq, xq_min), xq_max)

        nz_pairs = np.any([ xq[::2] != 0, xq[1::2] != 0 ], axis=0)
        lastnz = len(xq) - 2 * np.argmax(nz_pairs[-1::-1])
        if not np.any(nz_pairs):
            lastnz = 0

        return (xg, xq, lastnz)

    def compute_nbits(self, nbytes, x, lastnz, nbits_spec):

        mode = [ 0,   1 ][int(self.sr < T.SRATE_96K_HR and \
                              nbytes >= 20 * (3 + min(self.sr, T.SRATE_48K)))]
        rate = [ 0, 512 ][int(self.sr < T.SRATE_96K_HR and \
                              nbytes >  20 * (1 + min(self.sr, T.SRATE_48K)))]

        nbits_est = 0
        nbits_trunc = 0
        nbits_lsb = 0
        lastnz_trunc = 2
        c = 0

        for n in range(0, lastnz, 2):
            t = c + rate
            if n > len(x) // 2:
                t += 256

            a = abs(x[n  ])
            b = abs(x[n+1])
            lev = 0
            while max(a, b) >= 4:
                nbits_est += \
                    T.AC_SPEC_BITS[T.AC_SPEC_LOOKUP[t + lev*1024]][16]
                if lev == 0 and mode == 1:
                    nbits_lsb += 2
                else:
                    nbits_est += 2 * 2048

                a >>= 1
                b >>= 1
                lev = min(lev + 1, 3)

            nbits_est += \
                T.AC_SPEC_BITS[T.AC_SPEC_LOOKUP[t + lev*1024]][a + 4*b]

            a_lsb = abs(x[n  ])
            b_lsb = abs(x[n+1])
            nbits_est += (min(a_lsb, 1) + min(b_lsb, 1)) * 2048
            if lev > 0 and mode == 1:
                a_lsb >>= 1
                b_lsb >>= 1
                nbits_lsb += int(a_lsb == 0 and x[n  ] != 0)
                nbits_lsb += int(b_lsb == 0 and x[n+1] != 0)

            if (x[n] != 0 or x[n+1] != 0) and \
                    (nbits_est <= nbits_spec * 2048):
                lastnz_trunc = n + 2
                nbits_trunc = nbits_est

            t = 1 + (a + b) * (lev + 1) if lev <= 1 else 12 + lev
            c = (c & 15) * 16 + t

        nbits_est = (nbits_est + 2047) // 2048 + nbits_lsb
        nbits_trunc = (nbits_trunc + 2047) // 2048

        self.rate = rate
        self.lsb_mode = mode == 1 and nbits_est > nbits_spec

        return (nbits_est, nbits_trunc, lastnz_trunc, self.lsb_mode)

    def adjust_gain(self, g_idx, nbits, nbits_spec):

        T1 = [  80,  230,  380,  530,  680,  680,  830 ]
        T2 = [ 500, 1025, 1550, 2075, 2600, 2600, 3125 ]
        T3 = [ 850, 1700, 2550, 3400, 4250, 4250, 5100 ]

        dt = self.dt
        sr = self.sr

        if nbits < T1[sr]:
            delta = (nbits + 48) / 16

        elif nbits < T2[sr]:
            a = T1[sr] / 16 + 3
            b = T2[sr] / 48
            delta = a + (nbits - T1[sr]) * (b - a) / (T2[sr] - T1[sr])

        elif nbits < T3[sr]:
            delta = nbits / 48

        else:
            delta = T3[sr] / 48

        delta = np.fix(delta + 0.5).astype(int)

        if self.sr >= T.SRATE_48K_HR and \
            (g_idx < 255 and nbits > nbits_spec):

            factor = [ 3 + (nbits >= 520), 2, 0, 1 ][dt]
            g_incr = int(factor * (1 + (nbits - nbits_spec) / delta))
            return min(g_idx + g_incr, 255) - g_idx

        elif self.sr < T.SRATE_48K_HR and \
            ( (g_idx < 255 and nbits > nbits_spec) or \
              (g_idx >   0 and nbits < nbits_spec - (delta + 2)) ):

            if nbits < nbits_spec - (delta + 2):
                return -1

            if g_idx == 254 or nbits < nbits_spec + delta:
                return 1

            else:
                return 2

        return 0

    def estimate_noise(self, bw, xq, lastnz, x):

        i_nf = self.get_noise_indices(bw, xq, lastnz)
        l_nf = sum(abs(x[:len(i_nf)] * i_nf)) / sum(i_nf) \
            if sum(i_nf) > 0 else 0

        return min(max(np.rint(8 - 16 * l_nf).astype(int), 0), 7)

    def run(self, bw, nbytes, nbits_bw, nbits_ltpf, nbits_sns, nbits_tns, x):

        sr = self.sr

        ### Bit budget

        hr = self.sr >= T.SRATE_48K_HR

        nbits_gain = 8
        nbits_nf   = 3

        nbits_ari  = np.ceil(np.log2(len(x) / 2)).astype(int)
        nbits_ari += 3 + int(hr) + min((8*nbytes - 1) // 1280, 2)

        nbits_spec = 8*nbytes - \
            nbits_bw - nbits_ltpf - nbits_sns - nbits_tns - \
            nbits_gain - nbits_nf - nbits_ari

        ### Global gain estimation

        nbits_off = self.nbits_off + self.nbits_spec - self.nbits_est
        nbits_off = min(40, max(-40, nbits_off))

        nbits_off = 0 if self.reset_off else \
                    0.8 * self.nbits_off + 0.2 * nbits_off

        g_off = self.get_gain_offset(nbytes)

        (g_min, g_int, self.reset_off) = \
            self.estimate_gain(x, nbytes, nbits_spec, nbits_off, g_off)
        self.nbits_off = nbits_off
        self.nbits_spec = nbits_spec

        ### Quantization

        (xg, xq, lastnz) = self.quantize(g_int, x)

        (nbits_est, nbits_trunc, lastnz_trunc, _) = \
            self.compute_nbits(nbytes, xq, lastnz, nbits_spec)

        self.nbits_est = nbits_est

        ### Adjust gain and requantize

        g_adj = self.adjust_gain(g_int - g_off, nbits_est, nbits_spec)
        g_adj = max(g_int + g_adj, g_min + g_off) - g_int

        (xg, xq, lastnz) = self.quantize(g_adj, xg)

        (nbits_est, nbits_trunc, lastnz_trunc, lsb_mode) = \
            self.compute_nbits(nbytes, xq, lastnz, nbits_spec)

        self.g_idx = g_int + g_adj - g_off
        self.xq = xq
        self.lastnz = lastnz_trunc

        self.nbits_residual_max = nbits_spec - nbits_trunc + 4
        self.xg = xg

        ### Noise factor

        self.noise_factor = self.estimate_noise(bw, xq, lastnz, x)

        return (self.xq, self.lastnz, self.xg)

    def store(self, b):

        ne = T.I[self.dt][self.sr][-1]
        nbits_lastnz = np.ceil(np.log2(ne/2)).astype(int)

        b.write_uint((self.lastnz >> 1) - 1, nbits_lastnz)
        b.write_uint(self.lsb_mode, 1)
        b.write_uint(self.g_idx, 8)

    def encode(self, bits):

        ### Noise factor

        bits.write_uint(self.noise_factor, 3)

        ### Quantized data

        lsbs = []

        x = self.xq
        c = 0

        for n in range(0, self.lastnz, 2):
            t = c + self.rate
            if n > len(x) // 2:
                t += 256

            a = abs(x[n  ])
            b = abs(x[n+1])
            lev = 0
            while max(a, b) >= 4:

                bits.ac_encode(
                    T.AC_SPEC_CUMFREQ[T.AC_SPEC_LOOKUP[t + lev*1024]][16],
                    T.AC_SPEC_FREQ[T.AC_SPEC_LOOKUP[t + lev*1024]][16])

                if lev == 0 and self.lsb_mode:
                    lsb_0 = a & 1
                    lsb_1 = b & 1
                else:
                    bits.write_bit(a & 1)
                    bits.write_bit(b & 1)

                a >>= 1
                b >>= 1
                lev = min(lev + 1, 3)

            bits.ac_encode(
                T.AC_SPEC_CUMFREQ[T.AC_SPEC_LOOKUP[t + lev*1024]][a + 4*b],
                T.AC_SPEC_FREQ[T.AC_SPEC_LOOKUP[t + lev*1024]][a + 4*b])

            a_lsb = abs(x[n  ])
            b_lsb = abs(x[n+1])
            if lev > 0 and self.lsb_mode:
                a_lsb >>= 1
                b_lsb >>= 1

                lsbs.append(lsb_0)
                if a_lsb == 0 and x[n+0] != 0:
                    lsbs.append(int(x[n+0] < 0))

                lsbs.append(lsb_1)
                if b_lsb == 0 and x[n+1] != 0:
                    lsbs.append(int(x[n+1] < 0))

            if a_lsb > 0:
                bits.write_bit(int(x[n+0] < 0))

            if b_lsb > 0:
                bits.write_bit(int(x[n+1] < 0))

            t = 1 + (a + b) * (lev + 1) if lev <= 1 else 12 + lev
            c = (c & 15) * 16 + t

        ### Residual data

        if self.lsb_mode == 0:
            nbits_residual = min(bits.get_bits_left(), self.nbits_residual_max)

            for i in range(len(self.xg)):

                if self.xq[i] == 0:
                    continue

                bits.write_bit(self.xg[i] >= self.xq[i])
                nbits_residual -= 1
                if nbits_residual <= 0:
                    break

        else:
            nbits_residual = min(bits.get_bits_left(), len(lsbs))
            for lsb in lsbs[:nbits_residual]:
                bits.write_bit(lsb)


class SpectrumSynthesis(SpectrumQuantization):

    def __init__(self, dt, sr):

        super().__init__(dt, sr)

        (self.lastnz, self.lsb_mode, self.g_idx) = \
            (None, None, None)

    def fill_noise(self, bw, x, lastnz, f_nf, nf_seed):

        i_nf = self.get_noise_indices(bw, x, lastnz)

        k_nf = np.argwhere(i_nf)
        l_nf = (8 - f_nf)/16

        for k in k_nf:
            nf_seed = (13849 + nf_seed * 31821) & 0xffff
            x[k] = [ -l_nf, l_nf ][nf_seed < 0x8000]

        return x

    def load(self, b):

        ne = T.I[self.dt][self.sr][-1]
        nbits_lastnz = np.ceil(np.log2(ne/2)).astype(int)

        self.lastnz = (b.read_uint(nbits_lastnz) + 1) << 1
        self.lsb_mode = b.read_uint(1)
        self.g_idx = b.read_uint(8)

        if self.lastnz > ne:
            raise ValueError('Invalid count of coded samples')

    def decode(self, bits, bw, nbytes):

        ### Noise factor

        f_nf = bits.read_uint(3)

        ### Quantized data

        ne = T.I[self.dt][self.sr][-1]
        x  = np.zeros(ne)
        rate = [ 0, 512 ][int(self.sr < T.SRATE_96K_HR and \
                              nbytes >  20 * (1 + min(self.sr, T.SRATE_48K)))]

        levs = np.zeros(len(x), dtype=np.intc)
        c = 0

        for n in range(0, self.lastnz, 2):
            t = c + rate
            if n > len(x) // 2:
                t += 256

            for lev in range(14):

                s = t + min(lev, 3) * 1024

                sym = bits.ac_decode(
                    T.AC_SPEC_CUMFREQ[T.AC_SPEC_LOOKUP[s]],
                    T.AC_SPEC_FREQ[T.AC_SPEC_LOOKUP[s]])

                if sym < 16:
                    break

                if self.lsb_mode == 0 or lev > 0:
                    x[n  ] += bits.read_bit() << lev
                    x[n+1] += bits.read_bit() << lev

            if lev >= 14:
                raise ValueError('Out of range value')

            a = sym %  4
            b = sym // 4

            levs[n  ] = lev
            levs[n+1] = lev

            x[n  ] += a << lev
            x[n+1] += b << lev

            if x[n] and bits.read_bit():
                x[n] = -x[n]

            if x[n+1] and bits.read_bit():
                x[n+1] = -x[n+1]

            lev = min(lev, 3)
            t = 1 + (a + b) * (lev + 1) if lev <= 1 else 12 + lev
            c = (c & 15) * 16 + t

        ### Residual data

        nbits_residual = bits.get_bits_left()
        if nbits_residual < 0:
            raise ValueError('Out of bitstream')

        if self.lsb_mode == 0:

            xr = np.zeros(len(x), dtype=np.bool)

            for i in range(len(x)):

                if nbits_residual <= 0:
                    xr.resize(i)
                    break

                if x[i] == 0:
                    continue

                xr[i] = bits.read_bit()
                nbits_residual -= 1

        else:

            for i in range(len(levs)):

                if nbits_residual <= 0:
                    break

                if levs[i] <= 0:
                    continue

                lsb = bits.read_bit()
                nbits_residual -= 1
                if not lsb:
                    continue

                sign = int(x[i] < 0)

                if x[i] == 0:

                    if nbits_residual <= 0:
                        break

                    sign = bits.read_bit()
                    nbits_residual -= 1

                x[i] += [ 1, -1 ][sign]

        ### Set residual and noise

        nf_seed = sum(abs(x.astype(np.intc)) * range(len(x)))

        zero_frame = (self.lastnz <= 2 and x[0] == 0 and x[1] == 0
                      and self.g_idx <= 0 and f_nf >= 7)

        if self.lsb_mode == 0:

            for i in range(len(xr)):

                if x[i] and xr[i] == 0:
                    x[i] += [ -0.1875, -0.3125 ][x[i] < 0]
                elif x[i]:
                    x[i] += [  0.1875,  0.3125 ][x[i] > 0]

        if not zero_frame:
            x = self.fill_noise(bw, x, self.lastnz, f_nf, nf_seed)

        ### Rescale coefficients

        g_int = self.get_gain_offset(nbytes) + self.g_idx
        x *= 10 ** (g_int / 28)

        return x


def initial_state():
    return { 'nbits_off' : 0.0, 'nbits_spare' : 0 }


### ------------------------------------------------------------------------ ###

def check_estimate_gain(rng, dt, sr):

    ok = True

    analysis = SpectrumAnalysis(dt, sr)

    mismatch_count = 0
    for i in range(10):
        ne = T.I[dt][sr][-1]
        x  = rng.random(ne) * i * 1e2

        nbytes = 20 + int(rng.random() * 100)
        nbits_budget = 8 * nbytes - int(rng.random() * 100)
        nbits_off = rng.random() * 10
        g_off = 10 - int(rng.random() * 20)

        (_, g_int, reset_off) = \
            analysis.estimate_gain(x, nbytes, nbits_budget, nbits_off, g_off)

        (g_int_c, reset_off_c, _) = lc3.spec_estimate_gain(
            dt, sr, x, nbytes, nbits_budget, nbits_off, -g_off)

        if g_int_c != g_int:
            mismatch_count += 1

        ok = ok and (g_int_c == g_int or mismatch_count <= 1)
        ok = ok and (reset_off_c == reset_off or mismatch_count <= 1)

    return ok

def check_quantization(rng, dt, sr):

    ok = True

    analysis = SpectrumAnalysis(dt, sr)

    for g_int in range(-128, 128):

        ne = T.I[dt][sr][-1]
        x  = rng.random(ne) * 1e2
        nbytes = 20 + int(rng.random() * 30)

        (xg, xq, nq) = analysis.quantize(g_int, x)
        (xg_c, nq_c) = lc3.spec_quantize(dt, sr, g_int, x)

        ok = ok and np.amax(np.abs(1 - xg_c/xg)) < 1e-6
        ok = ok and nq_c == nq

    return ok

def check_compute_nbits(rng, dt, sr):

    ok = True

    analysis = SpectrumAnalysis(dt, sr)

    for nbytes in range(20, 150):

        nbits_budget = nbytes * 8 - int(rng.random() * 100)
        ne = T.I[dt][sr][-1]
        xq = (rng.random(ne) * 8).astype(int)
        nq = ne // 2 + int(rng.random() * ne // 2)

        nq = nq - nq % 2
        if xq[nq-2] == 0 and xq[nq-1] == 0:
            xq[nq-2] = 1

        (nbits, nbits_trunc, nq_trunc, lsb_mode) = \
            analysis.compute_nbits(nbytes, xq, nq, nbits_budget)

        (nbits_c, nq_c, _) = \
            lc3.spec_compute_nbits(dt, sr, nbytes, xq, nq, 0)

        (nbits_trunc_c, nq_trunc_c, lsb_mode_c) = \
            lc3.spec_compute_nbits(dt, sr, nbytes, xq, nq, nbits_budget)

        ok = ok and nbits_c == nbits
        ok = ok and nbits_trunc_c == nbits_trunc
        ok = ok and nq_trunc_c == nq_trunc
        ok = ok and lsb_mode_c == lsb_mode

    return ok

def check_adjust_gain(rng, dt, sr):

    ok = True

    analysis = SpectrumAnalysis(dt, sr)

    for g_idx in (0, 128, 254, 255):
        for nbits in range(50, 5000, 5):
            nbits_budget = int(nbits * (0.95 + (rng.random() * 0.1)))

            g_adj = analysis.adjust_gain(g_idx, nbits, nbits_budget)

            g_adj_c = lc3.spec_adjust_gain(
                dt, sr, g_idx, nbits, nbits_budget, 0)

            ok = ok and g_adj_c == g_adj

    return ok

def check_unit(rng, dt, sr):

    ok = True

    state_c = initial_state()

    bwdet = m_bwdet.BandwidthDetector(dt, sr)
    ltpf = m_ltpf.LtpfAnalysis(dt, sr)
    tns = m_tns.TnsAnalysis(dt)
    sns = m_sns.SnsAnalysis(dt, sr)
    analysis = SpectrumAnalysis(dt, sr)

    nbytes = 100

    for i in range(10):
        ns = T.NS[dt][sr]
        ne = T.I[dt][sr][-1]

        x = rng.random(ns) * 1e4
        e = rng.random(min(len(x), 64)) * 1e10

        if sr < T.SRATE_48K_HR:
            bwdet.run(e)
        pitch_present = ltpf.run(x)
        tns.run(x[:ne], sr, False, nbytes)
        sns.run(e, False, 0, x)

        (xq, nq, xg) = analysis.run(sr, nbytes,
            0 if sr >= T.SRATE_48K_HR else bwdet.get_nbits(),
            ltpf.get_nbits(), sns.get_nbits(), tns.get_nbits(), x[:ne])

        (xg_c, side_c) = lc3.spec_analyze(dt, sr,
            nbytes, pitch_present, tns.get_data(), state_c, x[:ne])

        ok = ok and side_c['g_idx'] == analysis.g_idx
        ok = ok and side_c['nq'] == nq
        ok = ok and np.amax(np.abs(1 - xg_c/xg)) < 1e-6

    return ok

def check_noise(rng, dt, bw, hrmode = False):

    ok = True

    analysis = SpectrumAnalysis(dt, bw)

    xq_off = [ 0.375, 0.5 ][hrmode]

    for i in range(10):
        ne = T.I[dt][bw][-1]
        xq = ((rng.random(ne) - 0.5) * 10 ** (0.5)).astype(int)
        nq = ne - int(rng.random() * 5)
        x  = xq - np.select([xq < 0, xq > 0], np.array([ xq_off, -xq_off ]))

        nf = analysis.estimate_noise(bw, xq, nq, x)
        nf_c = lc3.spec_estimate_noise(dt, bw, hrmode, x, nq)

        ok = ok and nf_c == nf

    return ok

def check_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_16K

    ok = True

    state_c = initial_state()

    for i in range(len(C.X_F[i0])):

        ne = T.I[dt][sr][-1]

        g_int = lc3.spec_estimate_gain(dt, sr, C.X_F[i0][i],
            0, C.NBITS_SPEC[i0][i], C.NBITS_OFFSET[i0][i], -C.GG_OFF[i0][i])[0]
        ok = ok and g_int == C.GG_IND[i0][i] + C.GG_OFF[i0][i]

        (x, nq) = lc3.spec_quantize(dt, sr,
            C.GG_IND[i0][i] + C.GG_OFF[i0][i], C.X_F[i0][i])
        x += np.select([x < 0, x > 0], np.array([ 0.375, -0.375 ]))
        ok = ok and np.any((np.trunc(x) - C.X_Q[i0][i]) == 0)
        ok = ok and nq == C.LASTNZ[i0][i]
        nbits = lc3.spec_compute_nbits(dt, sr,
            C.NBYTES[i0], C.X_Q[i0][i], C.LASTNZ[i0][i], 0)[0]
        ok = ok and nbits == C.NBITS_EST[i0][i]

        g_adj = lc3.spec_adjust_gain(dt, sr,
            C.GG_IND[i0][i], C.NBITS_EST[i0][i], C.NBITS_SPEC[i0][i], 0)
        ok = ok and g_adj == C.GG_IND_ADJ[i0][i] - C.GG_IND[i0][i]

        if C.GG_IND_ADJ[i0][i] != C.GG_IND[i0][i]:

            (x, nq) = lc3.spec_quantize(dt, sr,
                C.GG_IND_ADJ[i0][i] + C.GG_OFF[i0][i], C.X_F[i0][i])
            lastnz = C.LASTNZ_REQ[i0][i]
            x += np.select([x < 0, x > 0], np.array([ 0.375, -0.375 ]))
            ok = ok and np.any(((np.trunc(x) - C.X_Q_REQ[i0][i])[:lastnz]) == 0)

        tns_data = {
            'nfilters' : C.NUM_TNS_FILTERS[i0][i],
            'lpc_weighting' : [ True, True ],
            'rc_order' : [ C.RC_ORDER[i0][i][0], 0 ],
            'rc' : [ C.RC_I_1[i0][i] - 8, np.zeros(8, dtype = np.intc) ]
        }

        (x, side) = lc3.spec_analyze(dt, sr, C.NBYTES[i0],
            C.PITCH_PRESENT[i0][i], tns_data, state_c, C.X_F[i0][i])

        xq = x + np.select([x < 0, x > 0], np.array([ 0.375, -0.375 ]))
        xq = np.trunc(xq)

        ok = ok and np.abs(state_c['nbits_off'] - C.NBITS_OFFSET[i0][i]) < 1e-5
        if C.GG_IND_ADJ[i0][i] != C.GG_IND[i0][i]:
            xq = C.X_Q_REQ[i0][i]
            nq = C.LASTNZ_REQ[i0][i]
            ok = ok and side['g_idx'] == C.GG_IND_ADJ[i0][i]
            ok = ok and side['nq'] == nq
            ok = ok and np.any(((xq[:nq] - xq[:nq])) == 0)
        else:
            xq = C.X_Q[i0][i]
            nq = C.LASTNZ[i0][i]
            ok = ok and side['g_idx'] == C.GG_IND[i0][i]
            ok = ok and side['nq'] == nq
            ok = ok and np.any((xq[:nq] - C.X_Q[i0][i][:nq]) == 0)
        ok = ok and side['lsb_mode'] == C.LSB_MODE[i0][i]

        gg = C.GG[i0][i] if C.GG_IND_ADJ[i0][i] == C.GG_IND[i0][i] \
                else C.GG_ADJ[i0][i]

        nf = lc3.spec_estimate_noise(
                dt, C.P_BW[i0][i], False, C.X_F[i0][i] / gg, nq)
        ok = ok and nf == C.F_NF[i0][i]

    return ok

def check():

    rng = np.random.default_rng(1234)
    ok = True

    for dt in range(T.NUM_DT):
        for sr in range(T.SRATE_8K, T.SRATE_48K + 1):
            ok = ok and check_estimate_gain(rng, dt, sr)
            ok = ok and check_quantization(rng, dt, sr)
            ok = ok and check_compute_nbits(rng, dt, sr)
            ok = ok and check_adjust_gain(rng, dt, sr)
            ok = ok and check_unit(rng, dt, sr)
            ok = ok and check_noise(rng, dt, sr)

    for dt in ( T.DT_2M5, T.DT_5M, T.DT_10M ):
        for sr in ( T.SRATE_48K_HR, T.SRATE_96K_HR ):
            ok = ok and check_estimate_gain(rng, dt, sr)
            ok = ok and check_quantization(rng, dt, sr)
            ok = ok and check_compute_nbits(rng, dt, sr)
            ok = ok and check_adjust_gain(rng, dt, sr)
            ok = ok and check_unit(rng, dt, sr)
            ok = ok and check_noise(rng, dt, sr, True)

    for dt in ( T.DT_7M5, T.DT_10M ):
        ok = ok and check_appendix_c(dt)

    return ok

### ------------------------------------------------------------------------ ###



#
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

import lc3
import tables as T, appendix_c as C


### ------------------------------------------------------------------------ ###

class AttackDetector:

    def __init__(self, dt, sr):

        self.dt = dt
        self.sr = sr
        self.ms = T.DT_MS[dt]

        self.xn1 = 0
        self.xn2 = 0
        self.en1 = 0
        self.an1 = 0
        self.p_att = 0

    def is_enabled(self, nbytes):

        c1 = self.dt == T.DT_10M and \
             self.sr == T.SRATE_32K and nbytes > 80

        c2 = self.dt == T.DT_10M and \
             self.sr >= T.SRATE_48K and nbytes >= 100

        c3 = self.dt == T.DT_7M5 and \
             self.sr == T.SRATE_32K and nbytes >= 61 and nbytes < 150

        c4 = self.dt == T.DT_7M5 and \
             self.sr >= T.SRATE_48K and nbytes >= 75 and nbytes < 150

        return c1 or c2 or c3 or c4

    def run(self, nbytes, x):

        ### Downsampling and filtering input

        mf = int(16 * self.ms)

        r = len(x) // mf
        x_att = np.array([ np.sum(x[i*r:(i+1)*r]) for i in range(mf) ])

        x_hp = np.empty(mf)
        x_hp[0 ] = 0.375 * x_att[0 ] - 0.5 * self.xn1    + 0.125 * self.xn2
        x_hp[1 ] = 0.375 * x_att[1 ] - 0.5 * x_att[0   ] + 0.125 * self.xn1
        x_hp[2:] = 0.375 * x_att[2:] - 0.5 * x_att[1:-1] + 0.125 * x_att[0:-2]
        self.xn2 = x_att[-2]
        self.xn1 = x_att[-1]

        ### Energy calculation

        nb = int(self.ms / 2.5)

        e_att = np.array([ np.sum(np.square(x_hp[40*i:40*(i+1)]))
                           for i in range(nb) ])

        a_att = np.empty(nb)
        a_att[0] = np.maximum(0.25 * self.an1, self.en1)
        for i in range(1,nb):
            a_att[i] = np.maximum(0.25 * a_att[i-1], e_att[i-1])
        self.en1 = e_att[-1]
        self.an1 = a_att[-1]

        ### Attack Detection

        p_att = -1
        flags = [ (e_att[i] > 8.5 * a_att[i]) for i in range(nb) ]

        for (i, f) in enumerate(flags):
            if f: p_att = i

        f_att = p_att >= 0 or self.p_att - 1 >= nb // 2
        self.p_att = 1 + p_att

        return self.is_enabled(nbytes) and f_att


def initial_state():
    return { 'en1': 0.0, 'an1': 0.0, 'p_att': 0 }

### ------------------------------------------------------------------------ ###

def check_enabling(rng, dt):

    ok = True

    for sr in range(T.SRATE_8K, T.SRATE_48K + 1):

        attdet = AttackDetector(dt, sr)

        for nbytes in [ 61, 61-1, 75-1, 75, 80, 80+1, 100-1, 100, 150-1, 150 ]:

            f_att = lc3.attdet_run(dt, sr, nbytes,
                initial_state(), 2 * rng.random(T.NS[dt][sr]+6) - 1)

            ok = ok and f_att == attdet.is_enabled(nbytes)

    return ok

def check_unit(rng, dt, sr):

    ns = T.NS[dt][sr]
    ok = True

    attdet = AttackDetector(dt, sr)

    state_c = initial_state()
    x_c = np.zeros(ns+6)

    for run in range(100):

        ### Generate noise, and an attack at random point

        x = ((2 * rng.random(ns)) - 1) * (2 ** 8 - 1)
        x[(ns * rng.random()).astype(int)] *= 2 ** 7

        ### Check Implementation

        f_att = attdet.run(100, x)

        x_c = np.append(x_c[-6:], x)
        f_att_c = lc3.attdet_run(dt, sr, 100, state_c, x_c)

        ok = ok and f_att_c == f_att
        ok = ok and np.amax(np.abs(1 - state_c['en1']/attdet.en1)) < 2
        ok = ok and np.amax(np.abs(1 - state_c['an1']/attdet.an1)) < 2
        ok = ok and state_c['p_att'] == attdet.p_att

    return ok

def check_appendix_c(dt):

    i0 = dt - T.DT_7M5
    sr = T.SRATE_48K

    ok = True
    state = initial_state()

    x = np.append(np.zeros(6), C.X_PCM_ATT[i0][0])
    f_att = lc3.attdet_run(dt, sr, C.NBYTES_ATT[i0], state, x)
    ok = ok and f_att == C.F_ATT[i0][0]

    x = np.append(x[-6:], C.X_PCM_ATT[i0][1])
    f_att = lc3.attdet_run(dt, sr, C.NBYTES_ATT[i0], state, x)
    ok = ok and f_att == C.F_ATT[i0][1]

    return ok

def check():

    rng = np.random.default_rng(1234)
    ok = True

    for dt in range(T.NUM_DT):
        ok and check_enabling(rng, dt)

    for dt in ( T.DT_7M5, T.DT_10M ):
        for sr in ( T.SRATE_32K, T.SRATE_48K ):
            ok = ok and check_unit(rng, dt, sr)

    for dt in ( T.DT_7M5, T.DT_10M ):
        ok = ok and check_appendix_c(dt)

    return ok

### ------------------------------------------------------------------------ ###
